import pandas as pd
from pipa.common.logger import logger
from pipa.common.utils import get_timestamp
from pipa.common.config import OUTPUT_DIR
import sqlite3


class SQLiteConnector:
    def __init__(self, db_path):
        self.db_path = db_path

    def _connect(self):
        return sqlite3.connect(self.db_path)

    def fetch_table_as_dataframe(self, table_name):
        df = self.execute_query(f"SELECT * FROM {table_name}")
        logger.info(f"Table {table_name} fetched successfully.")
        return df

    def execute_query(self, query):
        with self._connect() as conn:
            df = pd.read_sql_query(query, conn)
        logger.info("Query executed successfully.")
        return df

    def execute_query_and_export_to_csv(self, query, output_filepath):
        df = self.execute_query(query)
        logger.info(f"Exporting query result to {output_filepath}")
        df.to_csv(output_filepath, index=False)
        return df

    def export_table_to_csv(self, table_name, output_filepath=None):
        output_filepath = output_filepath or f"{OUTPUT_DIR}/{table_name}.csv"
        df = self.fetch_table_as_dataframe(table_name)
        logger.info(f"Exporting table {table_name} to {output_filepath}")
        df.to_csv(output_filepath, index=False)
        return df

    def export_table_to_excel(self, table_name, output_filepath=None):
        output_filepath = output_filepath or f"{OUTPUT_DIR}/{table_name}.xlsx"
        df = self.fetch_table_as_dataframe(table_name)
        logger.info(f"Exporting table {table_name} to {output_filepath}")
        df.to_excel(output_filepath, index=False)
        return df


def export_dataframe_to_csv(filepath=None):
    """
    Decorator function that exports a pandas DataFrame to a CSV file.

    Args:
        filepath (str, optional): The output file path. If not provided, a default path will be used.

    Returns:
        The decorated function.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, pd.DataFrame):
                output_filepath = filepath or f"{OUTPUT_DIR}/pipa_{get_timestamp()}.csv"
                logger.info(f"Exporting DataFrame to {output_filepath}")
                result.to_csv(output_filepath, index=False)
            return result

        return wrapper

    return decorator



from typing import Optional
import pandas as pd
import json
from pipa.parser.perf_script_call import PerfScriptData
from typing import Dict


class Node:
    """
    Represents a node in the call graph.

    Attributes:
        addr (str): The address of the node.
        symbol (str): The symbol of the node.
        caller (str): The caller of the node.
        command (str | None, optional): The command associated with the node. Defaults to None.
        cycles (int, optional): The number of cycles. Defaults to 0.
        instructions (int, optional): The number of instructions. Defaults to 0.
    """

    def __init__(
        self,
        addr: str,
        symbol: str,
        caller: str,
        command: str | None = None,
        cycles: int = 0,
        instructions: int = 0,
    ):
        """
        Initialize a CallGraph Node object.

        Args:
            addr (str): The address of the call graph node.
            symbol (str): The symbol of the call graph node.
            caller (str): The caller of the call graph node.
            command (str | None, optional): The command associated with the call graph node. Defaults to None.
            cycles (int, optional): The number of cycles in the call graph node. Defaults to 0.
            instructions (int, optional): The number of instructions in the call graph node. Defaults to 0.
        """
        # Instruction Pointer
        self.addr = addr

        # function symbol
        self.symbol = symbol
        # example:
        # 1. Foam::DimensionedField<double, Foam::surfaceMesh>::operator+=+0xf4
        #   split from end to start
        _symbol_split = self.symbol.rsplit("+", maxsplit=1)
        self.function_name = _symbol_split[0]
        self.function_offset = (
            int(_symbol_split[1], 16) if len(_symbol_split) == 2 else None
        )

        self.caller = caller
        self.command = command
        self.cycles = cycles
        self.instructions = instructions
        self.events: Dict[str, int] = {}

    def get_event_value(self, event: str) -> int:
        return self.events[event] if event in self.events else 0

    def set_event_value(self, event: str, value: int):
        old = self.get_event_value(event)
        self.events[event] = value
        return old

    def add_event_value(self, event: str, value: int):
        old = self.get_event_value(event)
        self.events[event] = old + value

    def get_function_name(self):
        return self.function_name

    def get_offset(self) -> Optional[int]:
        return self.function_offset


class NodeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Node):
            return {
                "addr": obj.addr,
                "symbol": obj.symbol,
                "caller": obj.caller,
                "command": obj.command,
                "cycles": obj.cycles,
                "instructions": obj.instructions,
            }
        return super().default(obj)


class NodeTable:
    """
    A class representing a table of nodes.

    This class provides a dictionary-like interface for managing nodes.

    Attributes:
        _nodes (dict): The dictionary containing the nodes.

    Methods:
        __init__(self, nodes: dict | None = None): Initializes the NodeTable.
        __getitem__(self, key): Retrieves a node from the NodeTable.
        __setitem__(self, key, value): Sets a node in the NodeTable.
        __iter__(self): Returns an iterator for the NodeTable.
        __len__(self): Returns the number of nodes in the NodeTable.
        __str__(self): Returns a string representation of the NodeTable.
        __repr__(self): Returns a string representation of the NodeTable.
        __contains__(self, key): Checks if a node exists in the NodeTable.
        __delitem__(self, key): Deletes a node from the NodeTable.
        __add__(self, other): Returns a new NodeTable with nodes from both tables.
        __sub__(self, other): Returns a new NodeTable with nodes not present in the other table.
        __and__(self, other): Returns a new NodeTable with nodes present in both tables.
        __or__(self, other): Returns a new NodeTable with nodes from either table.
        __xor__(self, other): Returns a new NodeTable with nodes present in only one of the tables.
        __eq__(self, other): Checks if two NodeTables are equal.
        __ne__(self, other): Checks if two NodeTables are not equal.
        __lt__(self, other): Checks if one NodeTable is less than the other.
        __le__(self, other): Checks if one NodeTable is less than or equal to the other.
        __gt__(self, other): Checks if one NodeTable is greater than the other.
        __ge__(self, other): Checks if one NodeTable is greater than or equal to the other.
        from_perf_script_data(cls, perf_script: PerfScriptData, pids: list | None = None, cpus: list | None = None): Creates a NodeTable from a PerfScriptData object.
        from_perf_script_file(cls, perf_script_file: str, pids: list | None = None, cpus: list | None = None): Creates a NodeTable from a perf script file.
    """

    def __init__(self, nodes: Dict[str, Node] | None = None):
        """
        Initializes a node table object.

        Args:
            nodes (dict | None): A dictionary representing the nodes of the call graph.
                Defaults to None if not provided.

        Returns:
            None
        """
        self._nodes: Dict[str, Node] = nodes if nodes else {}

    def __getitem__(self, key):
        return self._nodes[key]

    def __setitem__(self, key, value):
        self._nodes[key] = value

    def __iter__(self):
        return iter(self._nodes)

    def __len__(self):
        return len(self._nodes)

    def __str__(self):
        return str(self._nodes)

    def __repr__(self):
        return repr(self._nodes)

    def __contains__(self, key):
        return key in self._nodes

    def __delitem__(self, key):
        del self._nodes[key]

    def __add__(self, other):
        return NodeTable({**self._nodes, **other._nodes})

    def __sub__(self, other):
        return NodeTable(
            {k: v for k, v in self._nodes.items() if k not in other._nodes}
        )

    def __and__(self, other):
        return NodeTable({k: v for k, v in self._nodes.items() if k in other._nodes})

    def __or__(self, other):
        return NodeTable({**self._nodes, **other._nodes})

    def __xor__(self, other):
        return NodeTable(
            {k: v for k, v in self._nodes.items() if k not in other._nodes}
            | {k: v for k, v in other._nodes.items() if k not in self._nodes}
        )

    def __eq__(self, other):
        return self._nodes == other._nodes

    def __ne__(self, other):
        return self._nodes != other._nodes

    def __lt__(self, other):
        return self._nodes < other._nodes

    def __le__(self, other):
        return self._nodes <= other._nodes

    def __gt__(self, other):
        return self._nodes > other._nodes

    def __ge__(self, other):
        return self._nodes >= other._nodes

    @classmethod
    def from_perf_script_data(
        cls,
        perf_script: PerfScriptData,
        pids: list | None = None,
        cpus: list | None = None,
    ):
        """
        Create a node table from PerfScriptData.

        Args:
            perf_script (PerfScriptData): The PerfScriptData object containing the performance script data.
            pids (list | None, optional): The PIDs to filter the data by. Defaults to None.
            cpus (list | None, optional): The CPUs to filter the data by. Defaults to None.

        Returns:
            cls: An instance of the class with the call graph created from the PerfScriptData.

        """
        if pids is not None:
            perf_script = perf_script.filter_by_pids(pids=pids)
        if cpus is not None:
            perf_script = perf_script.filter_by_cpus(cpus=cpus)

        res: Dict[str, Node] = {}
        for block in perf_script.blocks:
            header = block.header
            calls = block.calls

            if not calls:
                continue

            addr = calls[0].addr
            if addr in res:
                if header.event == "cycles":
                    res[addr].cycles += header.value
                elif header.event == "instructions":
                    res[addr].instructions += header.value
                else:
                    res[addr].add_event_value(event=header.event, value=header.value)
            else:
                res[addr] = Node(
                    addr=calls[0].addr,
                    symbol=calls[0].symbol,
                    caller=calls[0].caller,
                    command=header.command,
                    cycles=header.value if header.event == "cycles" else 0,
                    instructions=header.value if header.event == "instructions" else 0,
                )
                if header.event not in ["cycles", "instructions"]:
                    res[addr].set_event_value(event=header.event, value=header.value)
            for i in range(1, len(calls)):
                if calls[i].addr not in res:
                    res[calls[i].addr] = Node(
                        addr=calls[i].addr,
                        symbol=calls[i].symbol,
                        caller=calls[i].caller,
                    )

        return cls(nodes=res)

    def copy(self):
        return NodeTable({k: v for k, v in self._nodes.items()})

    @classmethod
    def from_perf_script_file(
        cls, perf_script_file: str, pids: list | None = None, cpus: list | None = None
    ):
        """
        Create a CallGraph instance from a perf script file.

        Args:
            perf_script_file (str): The path to the perf script file.

        Returns:
            CallGraph: The CallGraph instance created from the perf script file.
        """
        return cls.from_perf_script_data(
            PerfScriptData.from_file(perf_script_file), pids=pids, cpus=cpus
        )

    def to_dataframe(self):
        data = [
            {
                "addr": v.addr,
                "symbol": v.symbol,
                "caller": v.caller,
                "command": v.command,
                "cycles": v.cycles,
                "instructions": v.instructions,
                **v.events,
            }
            for v in self._nodes.values()
        ]
        return pd.DataFrame(data)



# pyelftools
from elftools.dwarf.dwarfinfo import DWARFInfo
from elftools.elf.elffile import ELFFile
from pipa.common.logger import logger
from pipa.common.utils import (
    tar,
    process_compression,
    FileFormat,
    check_file_format,
)
from pipa.parser.perf_buildid import PerfBuildidData
from pipa.common.cmd import run_command
from pipa.service.call_graph.addr import DEFAULT_BUILD_ID_DIR
from typing import List, Dict
from tempfile import mktemp, mkdtemp
import os


def find_all_source_files(dwarfinfo: DWARFInfo) -> List[str]:
    source_files = []
    # Iter all Compile Units, may be a source file or part of it.
    for CU in dwarfinfo.iter_CUs():
        # get the compile unit's line program (includes mapping from machine codes to souce codes)
        lineprog = dwarfinfo.line_program_for_CU(CU)
        if lineprog is None:
            continue
        top_die = CU.get_top_DIE()
        comp_dir = top_die.attributes.get("DW_AT_comp_dir")
        relative_dir: str = ""
        if comp_dir:
            if type(comp_dir) is str:
                relative_dir = comp_dir
            else:
                relative_dir = comp_dir.value.decode("utf-8")  # type: ignore
        delta = 1 if lineprog.header.version < 5 else 0

        for file_entry in lineprog["file_entry"]:
            file_name = file_entry.name.decode("utf-8")
            dir_index = file_entry.dir_index
            try:
                dir_name = lineprog["include_directory"][dir_index - delta].decode(
                    "utf-8"
                )
            except IndexError:
                dir_name = ""
            file_name = os.path.join(dir_name, file_name)
            if not os.path.exists(file_name):
                d = os.path.join(relative_dir, file_name)
                if not os.path.exists(d):
                    logger.warning(f"source file {d} couldn't found, please check")
                    continue
                else:
                    file_name = d
            realpath = os.path.realpath(file_name)
            source_files.append(realpath)
            if os.path.islink(file_name):
                abspath = os.path.abspath(file_name)
                if abspath != realpath:
                    source_files.append(abspath)
    return source_files


def get_archive_manifest(
    builid_data: PerfBuildidData, perf_buildid_dir=DEFAULT_BUILD_ID_DIR
) -> List[str]:
    manifest = []
    if not os.path.exists(perf_buildid_dir):
        logger.error(f"perf buildid dir {perf_buildid_dir} not exists!")
        return []
    perf_buildid_linkdir = os.path.realpath(perf_buildid_dir)
    for build_id in builid_data.buildid_lists.values():
        linkname = f".build-id/{build_id[0:2]}/{build_id[2:]}"
        linkfile = os.path.join(perf_buildid_dir, linkname)
        realfile = os.path.realpath(linkfile)
        manifest.append(linkname)
        manifest.append(os.path.relpath(realfile, perf_buildid_linkdir))
    return manifest


def archive(
    perf_data: str,
    output_path: str,
    replace_modules: Dict[str, str] = {},
):
    """PIPA Archive

    archive buildid and source files in a perf data

    Args:
        perf_data (str): perf data
        output_path (str): stored archive to output path
        replace_modules (Dict[str, str]): replace some modules' to specified one, useful when original module doesn't provide debuginfo at all. For example, '/path/to/workload/cmd' -> '/path/to/workload-debug/cmd', just pass replace_modules={'/path/to/workload/cmd': '/path/to/workload-debug/cmd'}
    """
    if not os.path.exists(perf_data):
        logger.error(f"perf data {perf_data} not exists!")
        return
    if not os.path.exists(output_path):
        os.makedirs(output_path)
        logger.warning(f"{output_path} not exists, created")
    perf_buildid = mktemp()
    run_command(
        f"perf buildid-list -i {perf_data} --with-hits | grep -v '^ ' > {perf_buildid}"
    )
    # get source files' locations
    perf_buildid_data = PerfBuildidData.from_file(perf_buildid)
    if len(perf_buildid_data.buildid_lists) == 0:
        logger.error("No buildid found in perf data")
        return
    logger.debug("Get original buildid list:")
    for bid in f"{perf_buildid_data.to_raw_dataframe()}".splitlines():
        logger.debug(bid)
    modules_list = perf_buildid_data.get_modules()
    # apply module replace
    for i in range(len(modules_list)):
        for o, v in replace_modules.items():
            m = modules_list[i]
            if m != o:
                continue
            modules_list[i] = v
    logger.debug("Final module list:")
    for i, m in enumerate(modules_list):
        logger.debug(f"{i}: {m}")
    source_files = []
    for module in modules_list:
        if not os.path.exists(module):
            logger.warning(f"Not found ELF File {module} in buildid list")
            continue
        # check if it's an elf file with debuginfo
        # if it's a compress file, extract to a tmpdir and will use the extracted elf file (if it contains) for further processing
        # if it's not a compress file or elf file, pass
        fformat = check_file_format(module)
        if fformat == FileFormat.xz:
            # buildid will generate a xz compressed file named like drm_vram_helper.ko.xz
            # it contains debuginfo elf, named like drm_vram_helper.ko
            tmpd = mkdtemp()
            extracted, _ = os.path.splitext(os.path.basename(module))
            extracted = os.path.join(tmpd, extracted)
            process_compression(
                compressed=module,
                decompressed=extracted,
                format=FileFormat.xz,
                decompress=True,
            )
            if not os.path.exists(extracted):
                logger.warning(
                    f"Extract {module} to {tmpd}, but expected elf file {extracted} not found"
                )
                continue
            module = extracted
        elif fformat != FileFormat.elf:
            continue
        # open elf file
        f = open(module, "rb")
        elffile = ELFFile(f)
        if not elffile.has_dwarf_info():
            logger.warning(f"{module} has no dwarf info, please provide debuginfo")
            f.close()
            continue
        # get dwarf info
        dwarfinfo = elffile.get_dwarf_info()
        if not dwarfinfo.has_debug_info:
            logger.warning(
                f"{module}'s dwarf lost debuginfo, check your compile methods"
            )
            continue
        source_files.extend(find_all_source_files(dwarfinfo=dwarfinfo))
        f.close()
    # get archive manifest
    archive_files = get_archive_manifest(perf_buildid_data)
    # generate archive
    buildid_tar = os.path.join(output_path, f"{perf_data}.buildid.tar")
    buildid_bz2 = f"{buildid_tar}.bz2"
    sourcefiles_tar = os.path.join(output_path, f"{perf_data}.sourcefiles.tar")
    sourcefiles_bz2 = f"{sourcefiles_tar}.bz2"
    tar(
        output_tar=buildid_tar,
        base_dir=DEFAULT_BUILD_ID_DIR,
        manifest=archive_files,
    )
    process_compression(
        decompressed=buildid_tar,
        compressed=buildid_bz2,
        format=FileFormat.bzip2,
        decompress=False,
    )
    tar(output_tar=sourcefiles_tar, manifest=source_files)
    process_compression(
        decompressed=sourcefiles_tar,
        compressed=sourcefiles_bz2,
        format=FileFormat.bzip2,
        decompress=False,
    )
    print(f"Created buildid archive: {buildid_bz2}")
    print(f"Created sourcefiles archive: {sourcefiles_bz2}")
    print("Now you can transfer them to another machine")
    print("Usage:")
    print(f"\t tar -xf {os.path.basename(buildid_bz2)} -C ~/.debug")
    print(f"\t tar -xf {os.path.basename(sourcefiles_bz2)} -C /")
    print("\t\t Or")
    print(
        f"\t tar -xf {os.path.basename(sourcefiles_bz2)} -C /path/to/source_file_prefix"
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Archive buildid and source files")
    parser.add_argument(
        "-i", "--perf-data", default="perf.data", help="Path to the perf data"
    )
    parser.add_argument("-o", "--output-path", default="./", help="Set output path")
    args = parser.parse_args()
    perf_data = args.perf_data
    output_path = args.output_path



from math import sqrt
from typing import Tuple, List, Optional, Set
from rich import print
from pipa.common.logger import logger
from enum import Enum, unique
from functools import partial
import tarfile
import os
import sys
import datetime
import shutil
import lzma
import bz2
import random


@unique
class FileFormat(Enum):
    xz = "xz"
    bzip2 = "bz2"
    elf = "elf"
    tar = "tar"
    other = "other"

    def __str__(self) -> str:
        return self.value


def check_file_format(file: str) -> FileFormat:
    """Check the file format of the given file.

    Args:
        file (str): file to be checked

    Returns:
        file_format: The file format of the given file.
    """
    with open(file, "rb") as f:
        # if omitted, will read until EOF
        # the slice is still valid
        magic = f.read(270)
        if magic[0:4] == b"\x7f\x45\x4c\x46":
            # elf file, 4 bytes
            return FileFormat.elf
        elif magic[0:6] == b"\xfd\x37\x7a\x58\x5a\x00":
            # .xz, 6 bytes
            return FileFormat.xz
        elif magic[0:3] == b"BZh":
            # .bz2, 4 bytes
            return FileFormat.bzip2
        elif magic[257:263] == b"ustar " or magic == b"gnutar":
            return FileFormat.tar
        return FileFormat.other


def tar(
    output_tar,
    manifest: List[str] | str,
    base_dir: Optional[str] = None,
):
    """Create a tar archive based on a manifest list.

    Args:
        output_tar (str): Path to the output tar file.
        base_dir (str): Base directory to use for relative paths.
        manifest_file (str | List[str]): file or files in tar archive.
    """
    with tarfile.open(output_tar, mode="w") as tar:
        if type(manifest) is str:
            manifest = [manifest]
        non_duplicate_lists: Set[Tuple[str, str]] = set()
        for file in manifest:
            full_path = file
            if base_dir:
                full_path = os.path.join(base_dir, full_path)
            if os.path.exists(full_path):
                # First add file to set to prevent duplicate files
                # Duplicate files will significantly increase the size of the tar archive
                non_duplicate_lists.add((full_path, file))
            else:
                logger.warning(
                    f"Warning: {full_path} does not exist and will be skipped."
                )
                continue
        # Add file to tar archive
        for f in non_duplicate_lists:
            full_path, file = f
            tar.add(full_path, arcname=file)


def untar(
    input_tar: str,
    output_dir: str,
):
    """
    Extract a tar archive to the specified directory.

    Args:
        input_tar (str): Path to the input tar file.
        output_dir (str): Directory to extract the files to.
    """
    with tarfile.open(input_tar, mode="r") as tar:
        tar.extractall(path=output_dir)
        logger.debug(f"Extracted {input_tar} to {output_dir}")


def process_compression(
    compressed: str, decompressed: str, format: FileFormat, decompress: bool = False
):
    """Compress or decompress a file based on the specified format.

    If in compress mode, will compress `decompressed` to `compressed`

    If in decompress mode, will decompress `compressed` to `decompressed`

    Args:
        compressed (str): Compressed file
        decompressed (str): Decompressed file
        format (file_format): Format of the compressed file
        decompress (bool, optional): Decompress mode. Defaults to False, which is compress mode.
    """
    if format == FileFormat.xz:
        compress_method = lzma.open
    elif format == FileFormat.bzip2:
        compress_method = bz2.open
    else:
        logger.warning(f"not support {format}'s compress or decompress")
        return
    f_in_f = (
        partial(compress_method, compressed)
        if decompress
        else partial(open, decompressed)
    )
    f_out_f = (
        partial(open, decompressed)
        if decompress
        else partial(compress_method, compressed)
    )
    with f_in_f("rb") as f_in, f_out_f("wb") as f_out:
        shutil.copyfileobj(f_in, f_out)


def find_closest_factor_pair(n: int) -> Tuple[int, int]:
    """
    Find closest factor pair of n

    Args:
        n (int): number to find

    Returns:
        Tuple[int, int]: (small factor A, large factor B), A * B = n
    """
    for i in range(int(sqrt(n)), 0, -1):
        if n % i == 0:
            return (i, n // i)


def get_timestamp():
    """
    Returns the current timestamp in the format "YYYY-MM-DD-HH-MM-SS".

    Returns:
        str: The current timestamp.
    """
    return datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")


def handle_user_cancelled(func):
    """
    Decorator function that handles user cancellation by catching the KeyboardInterrupt exception.

    Args:
        func: The function to be decorated.

    Returns:
        The decorated function.

    Raises:
        None.
    """

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt as e:
            # Print "Cancelled by user" and exit
            print("Cancelled by user")
            sys.exit(0)
        except TypeError as e:
            sys.exit(0)

    return wrapper


def generate_unique_rgb_color(data: List, generate_seed=True) -> Tuple[int, int, int]:
    """
    Generate unique RGB color from data

    Args:
        data (Tuple): Generate rgb from hash of data
        generate_seed (bool, optional): Generate random seed for hash. Defaults to True.

    Returns:
        Tuple[int, int, int]: rgb color, (r, g, b)
    """
    if generate_seed:
        data.append(random.randint(1, 256))
    # generate hash
    data_hash = hash(tuple(data))

    r = (data_hash & 0xFF0000) >> 16
    g = (data_hash & 0x00FF00) >> 8
    b = data_hash & 0x0000FF

    r = r % 256
    g = g % 256
    b = b % 256

    return (r, g, b)



from enum import Enum, unique
from typing import Optional


@unique
class SarDataIndex(Enum):
    Timestamp = "timestamp"
    Average = "Average:"
    CPUUtilsMetrics = {
        r"%usr": "float64",
        r"%nice": "float64",
        r"%sys": "float64",
        r"%iowait": "float64",
        r"%steal": "float64",
        r"%irq": "float64",
        r"%soft": "float64",
        r"%guest": "float64",
        r"%gnice": "float64",
        r"%idle": "float64",
    }
    CPUUtils = [Timestamp, "CPU", *CPUUtilsMetrics.keys()]
    AvgCPUUtils = [Average, "CPU", *CPUUtilsMetrics.keys()]
    CPUPressureStatsMetrics = {
        r"%scpu-10": "float64",
        r"%scpu-60": "float64",
        r"%scpu-300": "float64",
        r"%scpu": "float64",
    }
    CPUPressureStats = [Timestamp, *CPUPressureStatsMetrics.keys()]
    ProcessStatsMetrics = {"proc/s": "float64", "cswch/s": "float64"}
    ProcessStats = [Timestamp, *ProcessStatsMetrics.keys()]
    InterruptStatsMetrics = {"intr/s": "float64"}
    InterruptStats = [Timestamp, "INTR", *InterruptStatsMetrics.keys()]
    AvgInterruptStats = [Average, "INTR", *InterruptStatsMetrics.keys()]
    SwapStatsMetrics = {"pswpin/s": "float64", "pswpout/s": "float64"}
    SwapStats = [Timestamp, *SwapStatsMetrics.keys()]
    PagingStatsMetrics = {
        "pgpgin/s": "float64",
        "pgpgout/s": "float64",
        "fault/s": "float64",
        "majflt/s": "float64",
        "pgfree/s": "float64",
        "pgscank/s": "float64",
        "pgscand/s": "float64",
        "pgsteal/s": "float64",
        r"%vmeff": "float64",
    }
    PagingStats = [Timestamp, *PagingStatsMetrics.keys()]
    DiskIOStatsMetrics = {
        "tps": "float64",
        "rtps": "float64",
        "wtps": "float64",
        "dtps": "float64",
        "bread/s": "float64",
        "bwrtn/s": "float64",
        "bdscd/s": "float64",
    }
    DiskIOStats = [Timestamp, *DiskIOStatsMetrics.keys()]
    MemPressureStatsMetrics = {
        r"%smem-10": "float64",
        r"%smem-60": "float64",
        r"%smem-300": "float64",
        r"%smem": "float64",
        r"%fmem-10": "float64",
        r"%fmem-60": "float64",
        r"%fmem-300": "float64",
        r"%fmem": "float64",
    }
    MemPressureStats = [Timestamp, *MemPressureStatsMetrics.keys()]
    MemoryStatsMetrics = {
        "kbmemfree": "int",
        "kbavail": "int",
        "kbmemused": "int",
        r"%memused": "float64",
        "kbbuffers": "int",
        "kbcached": "int",
        "kbcommit": "int",
        r"%commit": "float64",
        "kbactive": "int",
        "kbinact": "int",
        "kbdirty": "int",
        "kbanonpg": "int",
        "kbslab": "int",
        "kbkstack": "int",
        "kbpgtbl": "int",
        "kbvmused": "int",
    }
    MemoryStats = [Timestamp, *MemoryStatsMetrics.keys()]
    SwapMemoryStatsMetrics = {
        "kbswpfree": "int",
        "kbswpused": "int",
        r"%swpused": "float",
        "kbswpcad": "int",
        r"%swpcad": "float",
    }
    SwapMemoryStats = [Timestamp, *SwapMemoryStatsMetrics.keys()]
    HugePagesStatsMetrics = {
        "kbhugfree": "int",
        "kbhugused": "int",
        r"%hugused": "float64",
        "kbhugrsvd": "int",
        "kbhugsurp": "int",
    }
    HugePagesStats = [Timestamp, *HugePagesStatsMetrics.keys()]
    FileSystemStatsMetrics = {
        "dentunusd": "int",
        "file-nr": "int",
        "inode-nr": "int",
        "pty-nr": "int",
    }
    FileSystemStats = [Timestamp, *FileSystemStatsMetrics.keys()]
    LoadStatsMetrics = {
        "runq-sz": "int",
        "plist-sz": "int",
        "ldavg-1": "float64",
        "ldavg-5": "float64",
        "ldavg-15": "float64",
        "blocked": "int",
    }
    LoadStats = [Timestamp, *LoadStatsMetrics.keys()]
    TTYStatsMetrics = {
        "rcvin/s": "float64",
        "xmtin/s": "float64",
        "framerr/s": "float64",
        "prtyerr/s": "float64",
        "brk/s": "float64",
        "ovrun/s": "float64",
    }
    TTYStats = [Timestamp, "TTY", *TTYStatsMetrics.keys()]
    IOPressureStatsMetrics = {
        r"%sio-10": "float64",
        r"%sio-60": "float64",
        r"%sio-300": "float64",
        r"%sio": "float64",
        r"%fio-10": "float64",
        r"%fio-60": "float64",
        r"%fio-300": "float64",
        r"%fio": "float64",
    }
    IOPressureStats = [Timestamp, *IOPressureStatsMetrics.keys()]
    DeviceIOStatsMetrics = {
        "tps": "float64",
        "rkB/s": "float64",
        "wkB/s": "float64",
        "dkB/s": "float64",
        "areq-sz": "float64",
        "aqu-sz": "float64",
        "await": "float64",
        r"%util": "float64",
    }
    DeviceIOStats = [Timestamp, "DEV", *DeviceIOStatsMetrics.keys()]
    NetUtilsMetrics = {
        "rxpck/s": "float64",
        "txpck/s": "float64",
        "rxkB/s": "float64",
        "txkB/s": "float64",
        "rxcmp/s": "float64",
        "txcmp/s": "float64",
        "rxmcst/s": "float64",
        r"%ifutil": "float64",
    }
    NetUtils = [Timestamp, "IFACE", *NetUtilsMetrics.keys()]
    NetErrorMetrics = {
        "rxerr/s": "float64",
        "txerr/s": "float64",
        "coll/s": "float64",
        "rxdrop/s": "float64",
        "txdrop/s": "float64",
        "txcarr/s": "float64",
        "rxfram/s": "float64",
        "rxfifo/s": "float64",
        "txfifo/s": "float64",
    }
    NetError = [Timestamp, "IFACE", *NetErrorMetrics.keys()]
    NFSClientStatsMetrics = {
        "call/s": "float64",
        "retrans/s": "float64",
        "read/s": "float64",
        "write/s": "float64",
        "access/s": "float64",
        "getatt/s": "float64",
    }
    NFSClientStats = [Timestamp, *NFSClientStatsMetrics.keys()]
    NFSServerStatsMetrics = {
        "scall/s": "float64",
        "badcall/s": "float64",
        "packet/s": "float64",
        "udp/s": "float64",
        "tcp/s": "float64",
        "hit/s": "float64",
        "miss/s": "float64",
        "sread/s": "float64",
        "swrite/s": "float64",
        "saccess/s": "float64",
        "sgetatt/s": "float64",
    }
    NFSServerStats = [Timestamp, *NFSServerStatsMetrics.keys()]
    SocketStatsMetrics = {
        "totsck": "int",
        "tcpsck": "int",
        "udpsck": "int",
        "rawsck": "int",
        "ip-frag": "int",
        "tcp-tw": "int",
    }
    SocketStats = [Timestamp, *SocketStatsMetrics.keys()]
    IPStatsMetrics = {
        "irec/s": "float64",
        "fwddgm/s": "float64",
        "idel/s": "float64",
        "orq/s": "float64",
        "asmrq/s": "float64",
        "asmok/s": "float64",
        "fragok/s": "float64",
        "fragcrt/s": "float64",
    }
    IPStats = [Timestamp, *IPStatsMetrics.keys()]
    IPErrorStatsMetrics = {
        "ihdrerr/s": "float64",
        "iadrerr/s": "float64",
        "iukwnpr/s": "float64",
        "idisc/s": "float64",
        "odisc/s": "float64",
        "onort/s": "float64",
        "asmf/s": "float64",
        "fragf/s": "float64",
    }
    IPErrorStats = [Timestamp, *IPErrorStatsMetrics.keys()]
    ICMPStatsMetrics = {
        "imsg/s": "float64",
        "omsg/s": "float64",
        "iech/s": "float64",
        "iechr/s": "float64",
        "oech/s": "float64",
        "oechr/s": "float64",
        "itm/s": "float64",
        "itmr/s": "float64",
        "otm/s": "float64",
        "otmr/s": "float64",
        "iadrmk/s": "float64",
        "iadrmkr/s": "float64",
        "oadrmk/s": "float64",
        "oadrmkr/s": "float64",
    }
    ICMPStats = [Timestamp, *ICMPStatsMetrics.keys()]
    ICMPErrorStatsMetrics = {
        "ierr/s": "float64",
        "oerr/s": "float64",
        "idstunr/s": "float64",
        "odstunr/s": "float64",
        "itmex/s": "float64",
        "otmex/s": "float64",
        "iparmpb/s": "float64",
        "oparmpb/s": "float64",
        "isrcq/s": "float64",
        "osrcq/s": "float64",
        "iredir/s": "float64",
        "oredir/s": "float64",
    }
    ICMPErrorStats = [Timestamp, *ICMPErrorStatsMetrics.keys()]
    TCPStatsMetrics = {
        "active/s": "float64",
        "passive/s": "float64",
        "iseg/s": "float64",
        "oseg/s": "float64",
    }
    TCPStats = [Timestamp, *TCPStatsMetrics.keys()]
    TCPExtStatsMetrics = {
        "atmptf/s": "float64",
        "estres/s": "float64",
        "retrans/s": "float64",
        "isegerr/s": "float64",
        "orsts/s": "float64",
    }
    TCPExtStats = [Timestamp, *TCPExtStatsMetrics.keys()]
    UDPStatsMetrics = {
        "idgm/s": "float64",
        "odgm/s": "float64",
        "noport/s": "float64",
        "idgmerr/s": "float64",
    }
    UDPStats = [Timestamp, *UDPStatsMetrics.keys()]
    IPv6SocketStatsMetrics = {
        "tcp6sck": "float64",
        "udp6sck": "float64",
        "raw6sck": "float64",
        "ip6-frag": "float64",
    }
    IPv6SocketStats = [Timestamp, *IPv6SocketStatsMetrics.keys()]
    IPv6StatsMetrics = {
        "irec6/s": "float64",
        "fwddgm6/s": "float64",
        "idel6/s": "float64",
        "orq6/s": "float64",
        "asmrq6/s": "float64",
        "asmok6/s": "float64",
        "imcpck6/s": "float64",
        "omcpck6/s": "float64",
        "fragok6/s": "float64",
        "fragcr6/s": "float64",
    }
    IPv6Stats = [Timestamp, *IPv6StatsMetrics.keys()]
    IPv6ErrorStatsMetrics = {
        "ihdrer6/s": "float64",
        "iadrer6/s": "float64",
        "iukwnp6/s": "float64",
        "i2big6/s": "float64",
        "idisc6/s": "float64",
        "odisc6/s": "float64",
        "inort6/s": "float64",
        "onort6/s": "float64",
        "asmf6/s": "float64",
        "fragf6/s": "float64",
        "itrpck6/s": "float64",
    }
    IPv6ErrorStats = [Timestamp, *IPv6ErrorStatsMetrics.keys()]
    ICMPv6StatsMetrics = {
        "imsg6/s": "float64",
        "omsg6/s": "float64",
        "iech6/s": "float64",
        "iechr6/s": "float64",
        "oechr6/s": "float64",
        "igmbq6/s": "float64",
        "igmbr6/s": "float64",
        "ogmbr6/s": "float64",
        "igmbrd6/s": "float64",
        "ogmbrd6/s": "float64",
        "irtsol6/s": "float64",
        "ortsol6/s": "float64",
        "irtad6/s": "float64",
        "inbsol6/s": "float64",
        "onbsol6/s": "float64",
        "inbad6/s": "float64",
        "onbad6/s": "float64",
    }
    ICMPv6Stats = [Timestamp, *ICMPv6StatsMetrics.keys()]
    ICMPv6ErrorStatsMetrics = {
        "ierr6/s": "float64",
        "idtunr6/s": "float64",
        "odtunr6/s": "float64",
        "itmex6/s": "float64",
        "otmex6/s": "float64",
        "iprmpb6/s": "float64",
        "oprmpb6/s": "float64",
        "iredir6/s": "float64",
        "oredir6/s": "float64",
        "ipck2b6/s": "float64",
        "opck2b6/s": "float64",
    }
    ICMPv6ErrorStats = [Timestamp, *ICMPv6ErrorStatsMetrics.keys()]
    UDPv6StatsMetrics = {
        "idgm6/s": "float64",
        "odgm6/s": "float64",
        "noport6/s": "float64",
        "idgmer6/s": "float64",
    }
    UDPv6Stats = [Timestamp, *UDPv6StatsMetrics.keys()]
    SoftNetStatsMetrics = {
        "total/s": "float64",
        "dropd/s": "float64",
        "squeezd/s": "float64",
        "rx_rps/s": "float64",
        "flw_lim/s": "float64",
    }
    SoftNetStats = [Timestamp, "CPU", *SoftNetStatsMetrics.keys()]
    AvgSoftNetStats = [Average, "CPU", *SoftNetStatsMetrics.keys()]
    CPUFreqMetrics = {"MHz": "float64"}
    CPUFreq = [Timestamp, "CPU", *CPUFreqMetrics.keys()]
    AvgCPUFreq = [Average, "CPU", *CPUFreqMetrics.keys()]
    TemperatureStatsMetrics = {"degC": "float64", r"%temp": "float64"}
    TemperatureStats = [Timestamp, "TEMP", *TemperatureStatsMetrics.keys(), "DEVICE"]
    BusStatsMetrics = {"maxpower": "int"}
    BusStats = [
        Timestamp,
        "BUS",
        "idvendor",
        "idprod",
        *BusStatsMetrics.keys(),
        "manufact",
        "product",
    ]
    FileSystemSpaceStatsMetrics = {
        "MBfsfree": "int",
        "MBfsused": "int",
        r"%fsused": "float64",
        r"%ufsused": "float64",
        "Ifree": "int",
        "Iused": "int",
        r"%Iused": "float64",
    }
    FileSystemSpaceStats = [
        Timestamp,
        *FileSystemSpaceStatsMetrics.keys(),
        "FILESYSTEM",
    ]

    @classmethod
    def contains(cls, item) -> Optional[Enum]:
        for k in cls:
            if item == k.value:
                return k
        return None

    @classmethod
    def avg_metric_to_all_metric(cls, item: Enum) -> Optional[Enum]:
        if item.name.startswith("Avg"):
            try:
                n = item.name.replace("Avg", "")
                return cls.__getitem__(n)
            except Exception:
                return None
        return None

    def __eq__(self, value: object) -> bool:
        return self.value == value

    def __hash__(self) -> int:
        return hash(self.name)



import pandas as pd
import re
from pipa.common.cmd import run_command
from pipa.common.hardware.cpu import NUM_CORES_PHYSICAL
from pipa.common.logger import logger
from typing import List
import multiprocessing


# Parser class, responsible for parsing SAR data
class SarParser:
    @staticmethod
    def parse_sar_bin_to_txt(sar_bin_path: str) -> List[str]:
        """
        Parses the SAR binary file into a list of lines.

        Args:
            sar_bin_path (str): Path to the SAR binary file.

        Returns:
            list: List of lines in the SAR binary file.
        """
        sar_lines = run_command(f"LC_ALL='C' sar -A -f {sar_bin_path}").split("\n")
        return sar_lines

    @staticmethod
    def split_sar_block(sar_lines: List[str]) -> List[List[str]]:
        """
        Splits the SAR block into individual blocks by '\n'.

        Args:
            sar_lines (list): List of SAR output lines.

        Returns:
            list: List of individual SAR blocks.
        """
        sar_lines = [l.strip() for l in sar_lines]
        return [
            list(filter(None, p.split("\n")))
            for p in "\n".join(sar_lines).split("\n\n")
        ]

    @staticmethod
    def trans_time_to_seconds(df: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms the timestamp column in the given DataFrame to seconds.

        Note this function sees each non-descending queue as a seperate day and raw timestamp format is %H:%M:%S.

        Thus this function can't deal with those interval more than 1 day, like: ["00:00:00", "00:00:00"].
        We see as the same time and parse it as ["1900-01-01 00:00:00", "1900-01-01 00:00:00"], but it also can be ["1900-01-01 00:00:00", "1900-01-02 00:00:00"].

        We are now try to use sadf instead of rewriting parsing sar file directly.

        Args:
            df (pandas.DataFrame): The DataFrame containing the timestamp column.

        Returns:
            pandas.DataFrame: The DataFrame with the timestamp column transformed to seconds.
        """
        day_prefix = 0
        result = []
        base_date = pd.Timestamp("1900-01-01")
        # iter all timestamp and add day prefix
        for i, ts in enumerate(df["timestamp"]):
            # switch to next day
            if i > 0 and ts < df["timestamp"].iloc[i - 1]:
                day_prefix += 1
            result.append(
                base_date + pd.Timedelta(days=day_prefix) + pd.to_timedelta(ts)
            )
        df["timestamp"] = result

        try:
            df["timestamp"] -= df.loc[:, "timestamp"].iloc[0]
            df["timestamp"] = df["timestamp"].dt.total_seconds()
        except IndexError as e:
            logger.warning(
                f"{df.columns.to_list()} column may has wrong format, please check the origin sar data"
            )
        return df

    @staticmethod
    def merge_one_line(sar_line: str) -> List[str]:
        """
        Merge a single line of SAR data into a list.

        Args:
            sar_line (str): The SAR data line to be merged.

        Returns:
            list: The merged SAR data as a list.
        """
        sar_line = sar_line.split()
        if sar_line[1] in ["AM", "PM"]:
            sar_line.pop(1)
        return sar_line

    @staticmethod
    def add_post_fix(sar_line: List[str], len_columns: int) -> List[str]:
        """
        Adds post-fix to the given SAR line to match the specified number of columns.

        Args:
            sar_line (list): The SAR line to add post-fix to.
            len_columns (int): The desired number of columns.

        Returns:
            list: The SAR line with post-fix added to match the specified number of columns.
        """
        while len(sar_line) < len_columns:
            sar_line.append("")
        if len(sar_line) > len_columns:
            sar_line[len_columns - 1] += " ".join(sar_line[len_columns:])
        return sar_line[:len_columns]

    @staticmethod
    def process_subtable(
        sar_columns: List[str],
        sar_blocks: List[str],
        processes_num=min(12, NUM_CORES_PHYSICAL),
    ) -> List[List[str]]:
        """
        Process the subtable data by merging lines and adding post-fixes.

        Args:
            sar_columns (list): List of SAR columns.
            sar_blocks (list): List of SAR blocks.
            processes_num (int, optional): Number of processes to use for parallel processing.
                Defaults to the minimum of 12 and the number of physical CPU cores.

        Returns:
            list: List of processed subtable data.

        """
        if len(sar_blocks) <= 10**6 or processes_num <= 1:
            # if the number of lines is less than 1e6, use single process
            return [
                SarParser.add_post_fix(SarParser.merge_one_line(x), len(sar_columns))
                for x in sar_blocks
            ]
        pool = multiprocessing.Pool(processes=processes_num)
        merged_lines = pool.map(SarParser.merge_one_line, sar_blocks)
        res = pool.starmap(
            SarParser.add_post_fix,
            zip(merged_lines, [len(sar_columns) for _ in range(len(merged_lines))]),
        )
        pool.close()
        pool.join()
        return res

    @staticmethod
    def sar_to_df(sar_blocks: List[str]) -> pd.DataFrame:
        """
        Convert SAR blocks to a pandas DataFrame.

        Args:
            sar_blocks (list): A list of SAR blocks.

        Returns:
            pandas.DataFrame: A DataFrame containing the processed SAR data.

        """
        while sar_blocks[0] == "":
            sar_blocks = sar_blocks[1:]

        time_pattern = r"\d{2}:\d{2}:\d{2}"
        sar_columns = sar_blocks[0].split()
        if re.match(time_pattern, sar_columns[0]):
            sar_columns = ["timestamp"] + sar_columns[1:]
        return pd.DataFrame(
            SarParser.process_subtable(sar_columns, sar_blocks[1:]),
            columns=sar_columns,
        )

    @staticmethod
    def parse_sar_bin(sar_bin_path: str) -> List[pd.DataFrame]:
        """
        Parses the SAR binary file and returns a list of dataframes.

        Args:
            sar_bin_path (str): The path to the SAR binary file.

        Returns:
            List[pd.DataFrame]: A list of dataframes containing the parsed SAR data.
        """
        sar_content = SarParser.parse_sar_bin_to_txt(sar_bin_path)
        return SarParser.parse_sar_string(sar_content)

    @staticmethod
    def parse_sar_txt(sar_txt_path: str) -> List[pd.DataFrame]:
        """
        Parses the SAR text file and returns a list of dataframes.

        Args:
            sar_txt_path (str): The path to the SAR text file.

        Returns:
            List[pd.DataFrame]: A list of dataframes containing the parsed SAR data.
        """
        with open(sar_txt_path, "r") as f:
            sar_content = f.readlines()
        return SarParser.parse_sar_string(sar_content)

    @staticmethod
    def parse_sar_string(sar_string: List[str]) -> List[pd.DataFrame]:
        """
        Parses the SAR string and returns a list of dataframes.

        Args:
            sar_string (str): The string containing the SAR data.

        Returns:
            List[pd.DataFrame]: A list of dataframes containing the parsed SAR data.
        """
        sar_data = SarParser.split_sar_block(sar_string)[1:]
        a = [SarParser.sar_to_df(d) for d in sar_data]
        l = 0
        res = []
        while l < len(a):  # merge dataframes with the same columns, use two pointers
            r = l + 1
            while r < len(a) and a[r].columns.equals(a[l].columns):
                r += 1
            res.append(pd.concat(a[l:r], axis=0))
            l = r
        return res



from .sar_index import SarDataIndex
import pandas as pd
from typing import Optional, Dict, List
from pipa.common.logger import logger


# Processor class, responsible for data processing
class SarProcessor:
    def __init__(self, sar_data: List[pd.DataFrame]):
        """
        Initialize the SarProcessor with parsed SAR data.

        Args:
            sar_data (List[pd.DataFrame]): List of DataFrames containing parsed SAR data.
        """
        self.sar_data = sar_data
        self.saridx_2_colidx: Dict[SarDataIndex, int] = {}
        for i, sard in enumerate(self.sar_data):
            scolumns = sard.columns.to_list()
            sindex = SarDataIndex.contains(scolumns)
            if sindex:
                self.saridx_2_colidx[sindex] = i
            else:
                logger.warning(
                    f"{scolumns} not supported in pipa sar parse, please report an issue"
                )
        for sindex in self.saridx_2_colidx.keys():
            all_m = SarDataIndex.avg_metric_to_all_metric(sindex)
            if all_m and all_m in self.saridx_2_colidx:
                all_m_i = self.saridx_2_colidx[all_m]
                sindex_i = self.saridx_2_colidx[sindex]
                avg_pd = self.sar_data[sindex_i]
                avg_pd = avg_pd.rename(columns={"Average:": "timestamp"})
                self.sar_data[all_m_i] = (
                    pd.concat([self.sar_data[all_m_i], avg_pd])
                    .drop_duplicates()
                    .reset_index(drop=True)
                )
                logger.debug(f"combine avg metric {sindex} to all metric {all_m}")

    def get_column_index(self, sar_index: SarDataIndex) -> Optional[int]:
        return self.saridx_2_colidx.get(sar_index)

    def filter_dataframe(self, df, data_type: str = "detail") -> pd.DataFrame:
        """
        Filters the given dataframe based on the specified data type.

        Parameters:
        - df: pandas.DataFrame
            The dataframe to be filtered.
        - data_type: str, optional
            The type of data to filter. Valid values are "detail", "raw", and "average".
            Defaults to "detail".

        Returns:
        - pandas.DataFrame
            The filtered dataframe.

        Raises:
        - ValueError: If an invalid data type is provided.
        """
        match data_type:
            case "detail":
                return df[df["timestamp"] != "Average:"]
            case "raw":
                return df
            case "average":
                return df[df["timestamp"] == "Average:"]
            case _:
                raise ValueError("Invalid type")

    def get_CPU_utilization(self, data_type: str = "detail") -> pd.DataFrame:
        """
        Retrieves the CPU utilization data from the SAR data.

        Args:
            data_type (str, optional): The type of CPU utilization data to retrieve. Defaults to "detail".
            Valid values are "detail", "raw", and "average".

        Returns:
            DataFrame: The filtered DataFrame containing the CPU utilization data.
        """
        idx = self.get_column_index(SarDataIndex.CPUUtils)
        if idx is None:
            raise KeyError(f"{SarDataIndex.CPUUtils} not found in sar data")
        util = self.filter_dataframe(self.sar_data[idx], data_type).astype(
            SarDataIndex.CPUUtilsMetrics.value
        )
        util[r"%util"] = 100 - util[r"%idle"]
        return util

    def get_CPU_util_avg_by_threads(self, threads: list = None) -> pd.DataFrame:
        """
        Retrieves the average CPU utilization detailed data for the specified threads.

        Args:
            threads (list): List of CPU threads to retrieve the utilization data for.
                            If None, retrieves the utilization data for all threads. Defaults to None.

        Returns:
            DataFrame: The filtered DataFrame containing the CPU utilization data for the specified threads.
        """
        util = self.get_CPU_utilization("average")
        util_threads = (
            util[util["CPU"].isin([str(t) for t in threads])]
            if threads
            else util[util["CPU"] == "all"]
        )
        return util_threads

    def get_CPU_util_avg_summary(self, threads: list = None) -> Dict:
        """
        Retrieves the average CPU utilization summary for the specified threads.

        Args:
            threads (list, optional): List of CPU threads to retrieve the utilization summary for. If None, retrieves
                                      the summary for all threads. Defaults to None.

        Returns:
            dict: A dictionary containing the average CPU utilization summary.
        """
        util_threads = self.get_CPU_util_avg_by_threads(threads)
        return util_threads.drop(columns=["timestamp", "CPU"]).mean().to_dict()

    def get_CPU_frequency(self, data_type: str = "detail") -> pd.DataFrame:
        """
        Returns the CPU frequency data.

        Args:
            data_type (str, optional): The type of CPU freqency data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the CPU frequency data.
        """
        idx = self.get_column_index(SarDataIndex.CPUFreq)
        if idx is None:
            raise KeyError(f"{SarDataIndex.CPUFreq} not found in sar data")
        return self.filter_dataframe(self.sar_data[idx], data_type).astype(
            {"MHz": "float64"}
        )

    def get_cpu_freq_avg(self, threads: list = None) -> Dict:
        """
        Returns the average CPU frequency data for the specified threads.

        Args:
            threads (list, optional): List of CPU threads to retrieve the frequency data for. If None, retrieves the
                                      frequency data for all threads. Defaults to None.

        Returns:
            dict: A dictionary containing the average CPU frequency data for the specified threads.
        """
        df = self.get_CPU_frequency("average")
        df = df[df["CPU"].isin([str(t) for t in threads])] if threads else df
        if df.empty:
            return {"cpu_frequency_mhz": 0}
        return {"cpu_frequency_mhz": df["MHz"].mean()}

    def get_CPU_util_freq(self, data_type: str = "detail") -> pd.DataFrame:
        """
        Returns the CPU utilization and frequency data.

        Args:
            data_type (str, optional): The type of CPU utilization and frequency data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the CPU utilization and frequency data.
        """
        util, freq = self.get_CPU_utilization(data_type), self.get_CPU_frequency(
            data_type
        )
        return pd.merge(util, freq, on=["timestamp", "CPU"])

    def get_network_statistics(
        self, data_type: str = "detail", on_failures: bool = False
    ) -> pd.DataFrame:
        """Get network statistics

        Args:
            data_type (str, optional): detail / raw / average. Defaults to "detail".
            on_failures (bool, optional): use NetError if true, else using NetStat. Defaults to False.

        Returns:
            pd.DataFrame: Dataframe containing the Network Stattistics data.
        """
        sar_loc = SarDataIndex.NetError if on_failures else SarDataIndex.NetUtils
        idx = self.get_column_index(sar_loc)
        if idx is None:
            raise KeyError(f"{sar_loc} not found in sar data")
        astype_t = (
            SarDataIndex.NetErrorMetrics.value
            if on_failures
            else SarDataIndex.NetUtilsMetrics.value
        )
        return self.filter_dataframe(self.sar_data[idx], data_type).astype(astype_t)

    def get_network_statistics_avg(self, on_failures: bool = False) -> List[Dict]:
        return (
            self.get_network_statistics(data_type="average", on_failures=on_failures)
            .drop(columns=["timestamp"])
            .to_dict(orient="records")
        )

    def get_memory_usage(self, data_type: str = "detail") -> pd.DataFrame:
        """
        Returns the memory usage data.

        Args:
            data_type (str, optional): The type of memory usage data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the memory usage data.
        """
        idx = self.get_column_index(SarDataIndex.MemoryStats)
        if idx is None:
            raise KeyError(f"{SarDataIndex.MemoryStats} not found in sar data")
        return self.filter_dataframe(self.sar_data[idx], data_type).astype(
            SarDataIndex.MemoryStatsMetrics.value
        )

    def get_memory_usage_avg(self) -> Dict:
        """
        Returns the average memory usage data.

        Returns:
            dict: A dictionary containing the average memory usage data.
        """
        return (
            self.get_memory_usage("average")
            .drop(columns=["timestamp"])
            .to_dict(orient="records")[0]
        )

    def get_disk_usage(
        self, dev: str = None, data_type: str = "detail"
    ) -> pd.DataFrame:
        """
        Returns the disk usage data.

        Args:
            data_type (str, optional): The type of disk usage data to retrieve. Defaults to "detail".

        Returns:
            pd.DataFrame: Dataframe containing the disk usage data.
        """
        idx = self.get_column_index(SarDataIndex.DeviceIOStats)
        if idx is None:
            raise KeyError(f"{SarDataIndex.DeviceIOStats} not found in sar data")
        df = self.filter_dataframe(self.sar_data[idx], data_type).astype(
            SarDataIndex.DeviceIOStatsMetrics.value
        )
        return df[df["DEV"] == dev] if dev else df

    def get_disk_usage_avg(self, dev: str = None) -> List[Dict] | Dict:
        """
        Returns the average disk usage data.

        Args:
            dev (str, optional): The disk device to retrieve the average data for. Defaults to None.

        Returns:
            list: A list of dictionaries containing the average disk usage data for each device.
            if dev is specified, returns a single dictionary.
        """
        disk_usage_avg = (
            self.get_disk_usage(dev, "average")
            .drop(columns=["timestamp"])
            .rename(columns={"%util": "%disk_util", "await": "disk_await"})
            .to_dict(orient="records")
        )
        return disk_usage_avg[0] if dev else disk_usage_avg



r"""
# What is PIPA?

PIPA (Progressive & Intelligent Performance Analytics) is a platform that aggregates a complete toolchain of performance data collection, processing, and analysis with
 advanced algorithms, enabling users to effortlessly obtain in - depth insights into the performance of their systems and applications. It bridges the gap between raw performance data and actionable information, allowing for quick identification of bottlenecks and optimization opportunities.

PIPA (枇杷, loquat) is a local fruit of Zhejiang, China. PIPA consists of three parts: loquat tree, flower and fruit, which represent the collecting & processing, analysis and conclusion of performance data respectively.
"""



import logging
import time
from functools import wraps
from typing import Optional

# Create a logger
logger = logging.getLogger(__name__)

# Set the logging level, it's the minimum acceptable level
logger.setLevel(logging.INFO)

# Create formatters with more context information
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(module)s:%(funcName)s:%(lineno)d - %(levelname)s - %(message)s"
)

# Create a stream handler and set its level
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.DEBUG)
stream_handler.setFormatter(formatter)

# Add file handler
file_handler = logging.FileHandler("pipa.log")
file_handler.setLevel(logging.DEBUG)
file_handler.setFormatter(formatter)

# Add handlers to logger
logger.addHandler(stream_handler)
logger.addHandler(file_handler)


def set_level(
    logger_level: Optional[str | int] = None, print_level: Optional[str | int] = None
):
    try:
        if logger_level:
            logger.setLevel(logger_level)
        if print_level:
            stream_handler.setLevel(print_level)
    except Exception as e:
        logger.warning(f"{e}")
        logger.warning(
            f"available logger levels: {','.join(logging.getLevelNamesMapping().keys())}"
        )
        logger.warning(f"use logger level: {logging.getLevelName(logger.level)}")
        logger.warning(f"use print level: {logging.getLevelName(stream_handler.level)}")


def log_execution(func):
    """Decorator to log function execution details"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(f"Executing function: {func.__name__}")
        start_time = time.time()
        try:
            result = func(*args, **kwargs)
            logger.info(f"Function {func.__name__} executed successfully")
            return result
        except Exception as e:
            logger.error(
                f"Function {func.__name__} failed with error: {str(e)}", exc_info=True
            )
            raise
        finally:
            end_time = time.time()
            logger.info(
                f"Function {func.__name__} execution time: {end_time - start_time:.4f} seconds"
            )

    return wrapper


if __name__ == "__main__":
    # Example usage of log_execution decorator
    @log_execution
    def test_function():
        logger.debug("Debug message")
        logger.info("Info message")
        time.sleep(0.5)
        return "Success"

    test_function()
    # Example usage
    logger.debug("This is a debug message")
    logger.info("This is an info message")
    logger.warning("This is a warning message")
    logger.error("This is an error message")
    logger.critical("This is a critical message")



from typing import Dict, List, Optional, Tuple

import os
import time

from pipa.common.logger import logger
from pipa.common.utils import find_closest_factor_pair

# multi processing
from multiprocessing.pool import Pool as PoolCls

# pyelftools
from elftools.dwarf.dwarfinfo import DWARFInfo
from elftools.elf.elffile import ELFFile
from elftools.dwarf.lineprogram import LineState, LineProgram
from elftools.elf.sections import SymbolTableSection

# capstone for disassemble
from capstone import Cs
from capstone import (
    CS_ARCH_ARM,
    CS_ARCH_ARM64,
    CS_MODE_32,
    CS_MODE_ARM,
    CS_MODE_V8,
    CS_ARCH_X86,
    CS_MODE_64,
    CS_ARCH_PPC,
)

DEFAULT_BUILD_ID_DIR = os.path.join(os.path.expanduser("~"), ".debug")
"""Default build-id dir"""

ADDR2LINE_OPT_STRENGTH = 1e9
"""Operation strength to indicate when to parallelize in addr2line"""


def get_arch_mode(elffile: ELFFile) -> Tuple[int, int]:
    """Judge arch and mode based on elf file

    Args:
        elffile (ELFFile): The elf file

    Raises:
        NotImplementedError: Will raise error when the arch is not supported

    Returns:
        Tuple[int, int]: (arch, mode)
    """
    arch = elffile["e_machine"]
    elf_header = elffile["e_ident"]

    if arch == "EM_X86_64":
        return (CS_ARCH_X86, CS_MODE_64)
    elif arch == "EM_386":
        return (CS_ARCH_X86, CS_MODE_32)
    elif arch == "EM_ARM":
        a = CS_ARCH_ARM
        if elf_header["EI_CLASS"] == "ELFCLASS32":
            return (a, CS_MODE_ARM)
        else:
            return (a, CS_MODE_ARM | CS_MODE_V8)
    elif arch == "EM_AARCH64":
        return (CS_ARCH_ARM64, CS_MODE_ARM)
    elif arch == "EM_PPC64":
        return (CS_ARCH_PPC, CS_MODE_64)
    else:
        raise NotImplementedError(
            f"Unsupported architecture. Arch: {arch} / ELF Header: {elf_header}"
        )


def get_symbol_addresses(
    elffile: ELFFile, func_name: Optional[str]
) -> Dict[str, Tuple[int, int]]:
    """Find all functions' name / start address / size in the given elf file

    Args:
        elffile (ELFFile): The elf file
        func_name (Optional[str]): If given, will only return the given function's info, otherwise return all functions' info

    Returns:
        tuple[str, Tuple[int, int]]: Take function's name as key, start address and size as value in tuple format: (start address, size)
    """
    symbol_addresses = {}
    for section in elffile.iter_sections():
        if not isinstance(section, SymbolTableSection):
            continue
        for symbol in section.iter_symbols():
            # function type
            if symbol["st_info"]["type"] == "STT_FUNC":
                name = symbol.name
                addr = symbol["st_value"]  # start address
                size = symbol["st_size"]  # function size
                if func_name is None or func_name == name:
                    symbol_addresses[name] = (addr, size)
    return symbol_addresses


def get_text_section(elffile: ELFFile) -> Tuple[bytes, int]:
    """Get data and start addr of text section in the elf file

    Args:
        elffile (ELFFile): The elf file

    Raises:
        ValueError: Not found `.text` section in the elf file

    Returns:
        Tuple[bytes, int]: (section data, section start address)
    """
    text_section = elffile.get_section_by_name(".text")
    if not text_section:
        raise ValueError("Not found .text section in the elf file")
    return text_section.data(), text_section["sh_addr"]


def disassemble_func(
    elfcodes: bytes, start_addr: int, arch: int, mode: int
) -> Dict[int, Tuple[str, str]]:
    """Disassemble function

    Args:
        elfcodes (bytes): The elf codes
        start_addr (int): Start address of assemble codes
        arch (int): The elf arch
        mode (int): The elf mode

    Returns:
        Dict[int, Tuple[str, str]]: The disassembled function, key is address, value is (mnemonic, op_str)
    """
    md = Cs(arch, mode)
    md.detail = False
    md.skipdata = True
    func_asm = {}
    for i in md.disasm(elfcodes, start_addr):
        # mnemonic indicates the instruction
        # op_str indicates the operands
        func_asm[i.address] = (i.mnemonic, i.op_str)
    return func_asm


def addr2lines(
    dwarfinfo: DWARFInfo,
    addresses: List[int],
    pool: PoolCls,
) -> List[Tuple[str, int, int, int, str]]:
    """Address to source line mapping

    Args:
        dwarfinfo (DWARFInfo): DWARFInfo
        addresses (List[int]): addresses to map
        pool (PoolCls): multiprocessing pool

    Returns:
        List[Tuple[str, int, int, int]]: The source line mapping, list of (file_name, address, line, column)

    Example:
    >>> import pipa.service.call_graph.addr as pipa_cfg
    >>> pipa_cfg.ADDR2LINE_OPT_STRENGTH = 2e8
    >>> dwarfinfo = DWARFInfo(elf)
    >>> with Pool(NUM_CORES_PHYSICAL) as pool:
    >>>     addr2lines(dwarfinfo, [0x400000, 0x400001], pool)
    [('main.c', 0x400000, 1, 1), ('main.c', 0x400001, 1, 2)]
    """
    operations_strength = ADDR2LINE_OPT_STRENGTH
    parallel = pool._processes  # type: ignore
    sourcelines: List[Tuple[str, int, int, int, str]] = []
    state_list: List[Tuple[int, LineState, LineState]] = []
    lineprog_list: List[Tuple[LineProgram, int, str]] = []
    address_len = len(addresses)
    if address_len <= 0:
        return sourcelines
    # Iter all Compile Units, may be a source file or part of it.
    for i, CU in enumerate(dwarfinfo.iter_CUs()):
        # get the compile unit's line program (includes mapping from machine codes to souce codes)
        lineprog = dwarfinfo.line_program_for_CU(CU)
        if lineprog is None:
            continue
        top_die = CU.get_top_DIE()
        comp_dir = top_die.attributes.get("DW_AT_comp_dir")
        relative_dir: str = ""
        if comp_dir:
            if type(comp_dir) is str:
                relative_dir = comp_dir
            else:
                relative_dir = comp_dir.value.decode("utf-8")  # type: ignore
        delta = 1 if lineprog.header.version < 5 else 0
        lineprog_list.append((lineprog, delta, relative_dir))
        prevstate = None

        # Iter all entries in the line program
        entries = lineprog.get_entries()
        for entry in entries:
            state = entry.state
            if state is None:
                continue
            if prevstate is not None and prevstate.address < state.address:
                state_list.append((i, prevstate, state))
            if state.end_sequence:
                prevstate = None
            else:
                prevstate = state
    # prepare for addr2line
    found_addresses = []
    state_len = len(state_list)
    operations_len = address_len * state_len
    logger.debug(f"\t addresses length: {address_len}")
    logger.debug(f"\t state tuple length: {state_len}")
    if operations_len < operations_strength:
        # sequential addr2line
        for state_tuple in state_list:
            for address in addresses:
                # Found address
                lineprog_index, prevstate, state = state_tuple
                lineprog, delta, relative_dir = lineprog_list[lineprog_index]
                # find nearest symbol
                if prevstate.address <= address < state.address:
                    file_entry = lineprog["file_entry"][prevstate.file - delta]
                    file_name = file_entry.name.decode("utf-8")
                    dir_index = file_entry.dir_index
                    try:
                        dir_name = lineprog["include_directory"][
                            dir_index - delta
                        ].decode("utf-8")
                    except IndexError as e:
                        logger.warning("\t Could not found dir name:")
                        logger.warning(f"\t\t address: {address}")
                        logger.warning(f"\t\t dir_index: {dir_index}")
                        logger.warning(f"\t\t delta: {delta}")
                        logger.warning(f"\t\t file_name: {file_name}")
                        logger.warning(f"\t\t line: {prevstate.line}")
                        logger.warning(f"\t\t column: {prevstate.column}")
                        dir_name = ""
                    file_name = os.path.join(dir_name, file_name)
                    sourcelines.append(
                        (
                            file_name,
                            address,
                            prevstate.line,
                            prevstate.column,
                            relative_dir,
                        )
                    )
                    found_addresses.append(address)
                    address_len -= 1
                if address_len <= 0:
                    return sourcelines
    else:
        logger.debug(
            f"\t {operations_len} >= {operations_strength}, will use parallel {parallel} processes to addr2line"
        )

        # allocate tasks
        stime = time.perf_counter()
        m, n = find_closest_factor_pair(parallel)
        state_p = [int(state_len / m)] * m
        for i in range(0, state_len % m):
            state_p[i] += 1
        address_p = [int(address_len / n)] * n
        for i in range(0, address_len % n):
            address_p[i] += 1
        allocate_task = []
        spi = 0
        for sp in state_p:
            api = 0
            for ap in address_p:
                allocate_task.append(
                    (state_list[spi : spi + sp], addresses[api : api + ap])
                )
                api += ap
            spi += sp
        assert len(allocate_task) == parallel
        etime = time.perf_counter()
        logger.debug(f"\t allocate tasks within {etime - stime} seconds")

        # process
        stime = time.perf_counter()
        results = pool.map(addr2l, allocate_task)
        etime = time.perf_counter()
        logger.debug(f"\t perform tasks within {etime - stime} seconds")

        # parse results
        for r in results:
            for rs in r:
                lp_index, addr, pvstate = rs
                lineprog, delta, relative_dir = lineprog_list[lp_index]
                file_entry = lineprog["file_entry"][pvstate.file - delta]
                file_name = file_entry.name.decode("utf-8")
                dir_index = file_entry.dir_index
                try:
                    dir_name = lineprog["include_directory"][dir_index - delta].decode(
                        "utf-8"
                    )
                except IndexError as e:
                    logger.warning("\t Could not found dir name:")
                    logger.warning(f"\t\t address: {addr}")
                    logger.warning(f"\t\t dir_index: {dir_index}")
                    logger.warning(f"\t\t delta: {delta}")
                    logger.warning(f"\t\t file_name: {file_name}")
                    logger.warning(f"\t\t line: {pvstate.line}")
                    logger.warning(f"\t\t column: {pvstate.column}")
                    dir_name = ""
                file_name = os.path.join(dir_name, file_name)
                sourcelines.append(
                    (file_name, addr, pvstate.line, pvstate.column, relative_dir)
                )
                found_addresses.append(addr)
                address_len -= 1
                if address_len <= 0:
                    return sourcelines
    not_found_addresses = list(filter(lambda x: x not in found_addresses, addresses))
    logger.warning(
        f"\t Not all address found sourcelines mapped: {not_found_addresses}"
    )
    logger.warning(f"\t\t {len(not_found_addresses)} addresses not found")
    logger.warning("\t\t check whether add source file info when compiling")
    return sourcelines


def addr2l(task: Tuple[List[Tuple[int, LineState, LineState]], List[int]]):
    """addr2line parallel worker

    Args:
        task (Tuple[List[Tuple[int, LineState, LineState]], List[int]]): the task to be processed, (state_list, addresses)

    Returns:
        result (List[Tuple[int, int, LineState]]): the result of addr2line
    """
    state_list, addresses = task
    result: List[Tuple[int, int, LineState]] = []
    for state_tuple in state_list:
        for address in addresses:
            # Found address
            lineprog_index, prevstate, state = state_tuple
            # find nearest symbol
            if prevstate.address <= address < state.address:
                result.append((lineprog_index, address, prevstate))
    return result



from .sar_parser import SarParser
from .sar_processor import SarProcessor
from .sar_plotter import SarPlotter


# Main class, coordinating parser, processor, and plotter
class SarData:
    def __init__(self, sar_string: str):
        """
        Initialize a SAR object with the given SAR string.

        Args:
            sar_string (str): The SAR string to parse.
        """
        parsed_data = SarParser.parse_sar_string(sar_string)
        self.processor = SarProcessor(parsed_data)
        self.plotter = SarPlotter(self.processor)

    def __getattr__(self, name):
        # Delegate methods to the processor or plotter
        if hasattr(self.processor, name):
            return getattr(self.processor, name)
        if hasattr(self.plotter, name):
            return getattr(self.plotter, name)
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )

    @classmethod
    def init_with_sar_txt(cls, sar_txt_path: str):
        """
        Initializes the SarData object using a SAR text file.

        Args:
            sar_txt_path (str): Path to the SAR text file.

        Returns:
            SarData: Initialized SarData object.
        """
        with open(sar_txt_path, "r") as f:
            sar_content = f.readlines()
        return cls(sar_content)

    @classmethod
    def init_with_sar_bin(cls, sar_bin_path: str):
        """
        Initializes the SarData object using a SAR binary file.

        Args:
            sar_bin_path (str): Path to the SAR binary file.

        Returns:
            SarData: Initialized SarData object.
        """
        sar_content = SarParser.parse_sar_bin_to_txt(sar_bin_path)
        return cls(sar_content)



import seaborn as sns
from pipa.common.logger import logger
from pipa.common.utils import generate_unique_rgb_color

from .sar_parser import SarParser
from .sar_processor import SarProcessor
from typing import Optional, List, Literal
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from pipa.parser import make_single_plot


# Plotter class, responsible for data visualization
class SarPlotter:
    def __init__(self, processor: SarProcessor):
        """
        Initialize the SarPlotter with a SarProcessor instance.

        Args:
            processor (SarProcessor): An instance of SarProcessor containing processed SAR data.
        """
        self.processor = processor

    def plot_interactive_CPU_metrics(
        self,
        threads: Optional[list[int]] = None,
        metrics: List[
            Literal[
                r"%usr",
                r"%nice",
                r"%sys",
                r"%iowait",
                r"%irq",
                r"%soft",
                r"%steal",
                r"%guest",
                r"%gnice",
                r"%idle",
                r"%util",
            ]
        ] = [r"%util"],
        aggregation: bool = False,
        raw_data: bool = False,
        show: bool = True,
        write_to_html: Optional[str] = None,
    ) -> List[go.Scatter] | go.Figure:
        """Plots interactive CPU metrics over time.

        You can generate your own plotly fig from returned scatters when raw_data is True. Else will generate figure and return it.

        Args:
            threads (Optional[list[int]], optional): Specify cpu to show in fig. Defaults to None.
                When in aggregation mode, none means all, otherwise aggregate select threads.
                When in non-aggregation mode, none means display all cpu threads, otherwise display selected threads.
            metrics (List[ Literal[ r, optional): The CPU metrics to show. Defaults to [r"%util"].
            aggregation (bool, optional): Whether to aggregate the data by CPU thread. Defaults to False.

        Returns:
            List[go.Scatter]: list of raw CPU metrics scatters.
        """
        df = self.processor.get_CPU_utilization()
        # minus 'all'
        cpu_counts = df["CPU"].nunique() - 1
        df = SarParser.trans_time_to_seconds(df)
        scatters = []
        if aggregation:
            if threads:
                df = df[df["CPU"].isin([str(t) for t in threads])]
                df = df.groupby("timestamp").mean(numeric_only=True).reset_index()
                df["CPU"] = "all"
            threads = ["all"]
        elif threads is None:
            threads = list(range(0, cpu_counts))
        for t in threads:
            cpu_data = df[df["CPU"] == str(t)]
            for i, y in enumerate(metrics):
                r, g, b = generate_unique_rgb_color([t, i], generate_seed=True)
                try:
                    scatters.append(
                        go.Scatter(
                            x=cpu_data["timestamp"],
                            y=cpu_data[y],
                            mode="lines+markers",
                            name=f"CPU {t} {y}",
                            # different colors
                            line=dict(color=f"rgb({r}, {g}, {b})"),
                        )
                    )
                except KeyError as e:
                    logger.warning(
                        f"metric {y} not found in columns {cpu_data.columns.to_list()}: {e}"
                    )
        if raw_data:
            return scatters
        else:
            return make_single_plot(
                scatters=scatters,
                title="CPU Metrics Trend",
                xaxis_title="Timestamp",
                yaxis_title="Percentage",
                show=show,
                write_to_html=write_to_html,
            )

    def plot_CPU_util_time(self, threads: list = None):
        """
        Plots the CPU utilization over time.

        Args:
            threads (list, optional): List of CPU threads to plot. If None, plots the utilization for all threads.
                                      Defaults to None.
        """
        df = self.processor.get_CPU_utilization()
        df = (
            df[df["CPU"].isin([str(t) for t in threads])]
            if threads
            else df.query("CPU=='all'")
        )
        df = SarParser.trans_time_to_seconds(df)

        if threads and len(threads) > 1:
            sns.lineplot(data=df, x="timestamp", y=r"%util", hue="CPU")
        else:
            sns.lineplot(data=df, x="timestamp", y=r"%util")

    def plot_interactive_CPU_freq(
        self,
        threads: Optional[list[int]] = None,
        aggregation: bool = False,
        raw_data: bool = False,
        show: bool = True,
        write_to_html: Optional[str] = None,
    ) -> List[go.Scatter] | go.Figure:
        """
        Plot raw scatters of CPU frequency metrics over time.

        You can generate the fig from returned scatters when raw_data is True. Else will generate figure and return it.

        Args:
            threads (Optional[list[int]], optional): CPU threads. Defaults to None, means choose 'all' CPU thread.

        Returns:
            List[go.Scatter]: list of raw scatter plots.
        """
        df = self.processor.get_CPU_frequency()
        # minus 'all'
        cpu_counts = df["CPU"].nunique() - 1
        df = SarParser.trans_time_to_seconds(df)
        scatters = []
        if aggregation:
            if threads:
                df = df[df["CPU"].isin([str(t) for t in threads])]
                df = df.groupby("timestamp").mean(numeric_only=True).reset_index()
                df["CPU"] = "all"
            threads = ["all"]
        elif threads is None:
            threads = list(range(0, cpu_counts))
        for t in threads:
            cpu_data = df[df["CPU"] == str(t)]
            r, g, b = generate_unique_rgb_color([t], generate_seed=True)
            scatters.append(
                go.Scatter(
                    x=cpu_data["timestamp"],
                    y=cpu_data["MHz"],
                    mode="lines+markers",
                    name=f"CPU {t} Freq",
                    # different colors
                    line=dict(color=f"rgb({r}, {g}, {b})"),
                )
            )
        if raw_data:
            return scatters
        else:
            return make_single_plot(
                scatters=scatters,
                title="CPU Freq Trend",
                xaxis_title="Timestamp",
                yaxis_title="MHz",
                show=show,
                write_to_html=write_to_html,
            )

    def plot_CPU_freq_time(self, threads: list = None):
        """
        Plots the CPU frequency over time.

        Args:
            threads (list, optional): List of CPU threads to plot. If None, plots the frequency for all threads.
                                      Defaults to None.
        """
        df = self.processor.get_CPU_frequency()

        sns.set_theme(style="darkgrid", rc={"figure.figsize": (15, 8)})

        df = (
            df[df["CPU"].isin([str(t) for t in threads])]
            if threads
            else df.query("CPU=='all'")
        )
        df = SarParser.trans_time_to_seconds(df)

        if threads and len(threads) > 1:
            sns.lineplot(
                data=df,
                x="timestamp",
                y="MHz",
                hue="CPU",
            )
        else:
            sns.lineplot(data=df, x="timestamp", y="MHz")

    def plot_interactive_network_stat(
        self,
        devs: list[str],
        trans_metrics: List[
            Literal[
                "rxpck/s",
                "txpck/s",
                "rxkB/s",
                "txkB/s",
                "rxcmp/s",
                "txcmp/s",
                "rxmcst/s",
                r"%ifutil",
            ]
        ] = [r"%ifutil"],
        err_metrics: List[
            Literal[
                "rxerr/s",
                "txerr/s",
                "coll/s",
                "rxdrop/s",
                "rxdrop/s",
                "txcarr/s",
                "rxfram/s",
                "rxfifo/s",
                "txfifo/s",
            ]
        ] = [r"rxerr/s"],
        on_failures=False,
        raw_data: bool = False,
        show: bool = True,
        write_to_html: Optional[str] = None,
    ) -> List[go.Scatter] | go.Figure:
        """
        Plots interactive network statistics over time.

        This function generates an interactive time series for the specified network devices and metrics.
        It can plot either transmission metrics or error metrics based on the `on_failures` flag.
        You can generate your own plotly fig from returned scatters when raw_data is True. Else will generate figure and return it.

        Args:
            devs (list[str]): A list of network device names to include in the plot.
            trans_metrics (List[Literal], optional): A list of transmission metrics to plot. Defaults to `["%ifutil"]`.
            err_metrics (List[Literal], optional): A list of error metrics to plot. Defaults to `["rxerr/s"]`.
            on_failures (bool, optional): If True, plots error metrics; otherwise, plots transmission metrics. Defaults to False.

        Returns:
            List[go.Scatter]: A list of Plotly Scatter objects representing the time series data for each device and metric.
        """
        if len(devs) < 1:
            return []
        metrics = err_metrics if on_failures else trans_metrics
        df = self.processor.get_network_statistics(on_failures=on_failures)
        df = SarParser.trans_time_to_seconds(df)

        scatters = []
        for t in devs:
            dev_data = df[df["IFACE"] == t]
            for i, y in enumerate(metrics):
                r, g, b = generate_unique_rgb_color([t, i], generate_seed=True)
                try:
                    scatters.append(
                        go.Scatter(
                            x=dev_data["timestamp"],
                            y=dev_data[y],
                            mode="lines+markers",
                            name=f"IFACE {t} {y}",
                            # different colors
                            line=dict(color=f"rgb({r}, {g}, {b})"),
                        )
                    )
                except KeyError as e:
                    logger.warning(
                        f"metric {y} not found in columns {dev_data.columns.to_list()}: {e}"
                    )
        if raw_data:
            return scatters
        else:
            return make_single_plot(
                scatters=scatters,
                title="Net Err Trend" if on_failures else "Net Stat Trend",
                xaxis_title="Timestamp",
                yaxis_title="Net Stat",
                show=show,
                write_to_html=write_to_html,
            )

    def plot_interactive_mem_usage(
        self,
        metrics: List[
            Literal[
                "kbmemfree",
                "kbavail",
                "kbmemused",
                r"%memused",
                "kbbuffers",
                "kbcached",
                "kbcommit",
                r"%commit",
                "kbactive",
                "kbinact",
                "kbdirty",
                "kbanonpg",
                "kbslab",
                "kbkstack",
                "kbpgtbl",
                "kbvmused",
            ]
        ] = [r"%memused"],
        raw_data: bool = False,
        show: bool = True,
        write_to_html: Optional[str] = None,
    ) -> List[go.Scatter] | go.Figure:
        """
        Generates interactive memory usage time series plots.

        This function generates the time series data for the specified memory metrics.
        Each metric is plotted with a unique color.
        You can generate your own plotly fig from returned scatters when raw_data is True. Else will generate figure and return it.

        Args:
            metrics (List[Literal], optional): A list of memory metrics to plot. Defaults to `["%memused"]`.

        Returns:
            List[go.Scatter]: A list of Plotly Scatter objects representing the time series data for each memory metric.
        """
        df = self.processor.get_memory_usage()
        df = SarParser.trans_time_to_seconds(df)

        scatters = []
        for i, y in enumerate(metrics):
            r, g, b = generate_unique_rgb_color([i], generate_seed=True)
            try:
                scatters.append(
                    go.Scatter(
                        x=df["timestamp"],
                        y=df[y],
                        mode="lines+markers",
                        name=f"memory {y}",
                        # different colors
                        line=dict(color=f"rgb({r}, {g}, {b})"),
                    )
                )
            except KeyError as e:
                logger.warning(
                    f"metric {y} not found in columns {df.columns.to_list()}: {e}"
                )
        if raw_data:
            return scatters
        else:
            return make_single_plot(
                scatters=scatters,
                title="Memory Metrics Trend",
                xaxis_title="Timestamp",
                yaxis_title="Memory Usage",
                show=show,
                write_to_html=write_to_html,
            )

    def plot_memory_usage(self):
        """
        Plots the memory usage over time.
        """
        df = self.processor.get_memory_usage()
        df = SarParser.trans_time_to_seconds(df)
        sns.lineplot(data=df, x="timestamp", y=r"%memused")

    def plot_interactive_disk_usage(
        self,
        devs: list[str],
        metrics: List[
            Literal[
                "tps",
                r"rkB/s",
                r"wkB/s",
                r"dkB/s",
                "areq-sz",
                "aqu-sz",
                "await",
                r"%util",
            ]
        ] = [r"%util"],
        raw_data: bool = False,
        show: bool = True,
        write_to_html: Optional[str] = None,
    ) -> List[go.Scatter] | go.Figure:
        """
        Generates interactive disk usage time series plots.

        This function generates the time series data for the specified disk devices and metrics.
        Each metric for each device is plotted with a unique color.
        You can generate your own plotly fig from returned scatters when raw_data is True. Else will generate figure and return it.

        Args:
            devs (list[str]): A list of disk device names to include in the plot.
            metrics (List[Literal], optional): A list of disk usage metrics to plot. Defaults to `["%util"]`.

        Returns:
            List[go.Scatter]: A list of Plotly Scatter objects representing the time series data for each disk device and metric.
        """
        if len(devs) < 1:
            return []
        df = self.processor.get_disk_usage()
        df = SarParser.trans_time_to_seconds(df)

        scatters = []
        for t in devs:
            cpu_data = df[df["DEV"] == t]
            for i, y in enumerate(metrics):
                r, g, b = generate_unique_rgb_color([t, i], generate_seed=True)
                try:
                    scatters.append(
                        go.Scatter(
                            x=cpu_data["timestamp"],
                            y=cpu_data[y],
                            mode="lines+markers",
                            name=f"DEV {t} {y}",
                            # different colors
                            line=dict(color=f"rgb({r}, {g}, {b})"),
                        )
                    )
                except KeyError as e:
                    logger.warning(
                        f"metric {y} not found in columns {cpu_data.columns.to_list()}: {e}"
                    )
        if raw_data:
            return scatters
        else:
            return make_single_plot(
                scatters=scatters,
                title="Disk Usage Trend",
                xaxis_title="Timestamp",
                yaxis_title="Disk Usage",
                show=show,
                write_to_html=write_to_html,
            )

    def plot_disk_usage(self, dev: str = None, metrics="tps"):
        """
        Plots the disk tps over time.
        """
        df = self.processor.get_disk_usage()
        df = SarParser.trans_time_to_seconds(df)
        df = df.query(f"DEV=='{dev}'") if dev else df
        if dev:
            sns.lineplot(data=df, x="timestamp", y=metrics)
        else:
            sns.lineplot(data=df, x="timestamp", y=metrics, hue="DEV")

    def plot_all_metrics(
        self,
        net_devs: list[str],
        disk_devs: list[str],
        cpu_threads: Optional[list[int]] = None,
        cpu_metrics: List[
            Literal[
                r"%usr",
                r"%nice",
                r"%sys",
                r"%iowait",
                r"%irq",
                r"%soft",
                r"%steal",
                r"%guest",
                r"%gnice",
                r"%idle",
                r"%util",
            ]
        ] = [r"%util"],
        cpu_aggregation: bool = False,
        net_trans_metrics: List[
            Literal[
                "rxpck/s",
                "txpck/s",
                "rxkB/s",
                "txkB/s",
                "rxcmp/s",
                "txcmp/s",
                "rxmcst/s",
                r"%ifutil",
            ]
        ] = [r"%ifutil"],
        net_err_metrics: List[
            Literal[
                "rxerr/s",
                "txerr/s",
                "coll/s",
                "rxdrop/s",
                "rxdrop/s",
                "txcarr/s",
                "rxfram/s",
                "rxfifo/s",
                "txfifo/s",
            ]
        ] = [r"rxerr/s"],
        mem_metrics: List[
            Literal[
                "kbmemfree",
                "kbavail",
                "kbmemused",
                r"%memused",
                "kbbuffers",
                "kbcached",
                "kbcommit",
                r"%commit",
                "kbactive",
                "kbinact",
                "kbdirty",
                "kbanonpg",
                "kbslab",
                "kbkstack",
                "kbpgtbl",
                "kbvmused",
            ]
        ] = [r"%memused"],
        disk_metrics: List[
            Literal[
                "tps",
                r"rkB/s",
                r"wkB/s",
                r"dkB/s",
                "areq-sz",
                "aqu-sz",
                "await",
                r"%util",
            ]
        ] = [r"%util"],
        show: bool = True,
        write_to_html: Optional[str] = None,
        height=1000,
        shared_xaxes=True,
        vertical_spacing=0.1,
    ) -> go.Figure:
        """
        Plots comprehensive system metrics including CPU utilization, CPU frequency, network transmission, network errors, memory usage, and disk usage.

        This function generates an interactive plot with multiple subplots, each representing a different system metric. The plot can be displayed and optionally saved as an HTML file.

        Args:
            net_devs (list[str]): A list of network device names to include in the plot.
            disk_devs (list[str]): A list of disk device names to include in the plot.
            cpu_threads (Optional[list[int]], optional): A list of CPU thread IDs to include in the CPU metrics. Defaults to None.
            cpu_metrics (List[Literal], optional): A list of CPU metrics to plot. Defaults to `["%util"]`.
            cpu_aggregation (bool, optional): Whether to aggregate CPU metrics across all threads. Defaults to False.
            net_trans_metrics (List[Literal], optional): A list of network transmission metrics to plot. Defaults to `["%ifutil"]`.
            net_err_metrics (List[Literal], optional): A list of network error metrics to plot. Defaults to `["rxerr/s"]`.
            mem_metrics (List[Literal], optional): A list of memory metrics to plot. Defaults to `["%memused"]`.
            disk_metrics (List[Literal], optional): A list of disk usage metrics to plot. Defaults to `["%util"]`.
            show (bool, optional): Whether to show the plot. Defaults to True
            write_to_html (Optional[str], optional): The filename to save the plot as an HTML file. If None, the plot is not saved. Defaults to None.
            height (int, optional): The height of the plot in pixels. Defaults to 1000.
            shared_xaxes (bool, optional): Whether to share the x-axis across subplots. Defaults to True.
            vertical_spacing (float, optional): The vertical spacing between subplots. Defaults to 0.1.
        """
        cpu_util_scatters = self.plot_interactive_CPU_metrics(
            threads=cpu_threads,
            metrics=cpu_metrics,
            aggregation=cpu_aggregation,
            raw_data=True,
        )
        cpu_freq_scatters = self.plot_interactive_CPU_freq(
            threads=cpu_threads, raw_data=True
        )
        net_trans_scatters = self.plot_interactive_network_stat(
            on_failures=False,
            devs=net_devs,
            trans_metrics=net_trans_metrics,
            raw_data=True,
        )
        net_err_scatters = self.plot_interactive_network_stat(
            on_failures=True, devs=net_devs, err_metrics=net_err_metrics, raw_data=True
        )
        mem_scatters = self.plot_interactive_mem_usage(
            metrics=mem_metrics, raw_data=True
        )
        disk_scatters = self.plot_interactive_disk_usage(
            devs=disk_devs, metrics=disk_metrics, raw_data=True
        )
        # subtitle: scatters, x_title, y_title
        all_scatters = {
            "CPU Utilization": (cpu_util_scatters, "timestamp", "Percentage"),
            "CPU Frequency": (cpu_freq_scatters, "timestamp", "MHz"),
            "Network Transmission": (net_trans_scatters, "timestamp", "Net Stat"),
            "Network Error": (net_err_scatters, "timestamp", "Net Stat"),
            "Memory Usage": (mem_scatters, "timestamp", "Memory Usage"),
            "Disk Usage": (disk_scatters, "timestamp", "Disk Usage"),
        }
        rows = 0
        sub_titles = []
        exist_scatters = {}
        for k, v in all_scatters.items():
            s, _, _ = v
            if len(s) > 0:
                rows += 1
                sub_titles.append(k)
                # scatters, x_title, y_title
                exist_scatters[k] = v
        fig = make_subplots(
            rows=rows,
            cols=1,
            subplot_titles=sub_titles,
            # Share same x axis since all is timestamp
            shared_xaxes=shared_xaxes,
            vertical_spacing=vertical_spacing,
        )
        for i, (_, v) in enumerate(exist_scatters.items()):
            s, xt, yt = v
            for scatter in s:
                fig.add_trace(scatter, row=i + 1, col=1)
            fig.update_xaxes(title_text=xt, row=i + 1, col=1)
            fig.update_yaxes(title_text=yt, row=i + 1, col=1)
        fig.update_layout(
            title="System Metrics Trends",
            hovermode="closest",
            height=height,
            showlegend=True,
        )
        if show:
            fig.show()
        if write_to_html:
            fig.write_html(write_to_html)
        return fig



from typing import Literal, Mapping, Optional
import networkx as nx
import json
import matplotlib.pyplot as plt
from pipa.parser.perf_buildid import PerfBuildidData
from pipa.parser.perf_script_call import PerfScriptData
from networkx.drawing.nx_pydot import write_dot
from collections import defaultdict
from pipa.service.call_graph.node import NodeTable
from pipa.service.call_graph.func_node import FunctionNodeTable, ClusterEncoder


class CallGraph:
    """
    Represents a call graph.

    Attributes:
        block_graph (nx.DiGraph): The directed graph representing the call relationships at the block level.
        node_table (NodeTable): The table mapping node addresses to node objects.
        func_graph (nx.DiGraph): The directed graph representing the call relationships at the function level.
        function_node_table (FunctionNodeTable): The table mapping function names to function nodes.
    """

    def __init__(
        self,
        block_graph: nx.DiGraph | None = None,
        node_table: NodeTable | None = None,
        func_graph: nx.DiGraph | None = None,
        function_node_table: FunctionNodeTable | None = None,
    ):
        """
        Initializes a CallGraph object.

        Args:
            block_graph (nx.DiGraph, optional): The directed graph representing the call relationships at blocks level.
                Defaults to None.
            node_table (NodeTable, optional): The table mapping node addresses to node objects.
                Defaults to None.
            func_graph (nx.DiGraph, optional): The directed graph representing the call relationships at function level.
                Defaults to None.
            function_node_table (FunctionNodeTable, optional): The table mapping function names to function nodes.
                Defaults to None.
        """
        self.block_graph = nx.DiGraph() if block_graph is None else block_graph
        self.node_table = NodeTable() if node_table is None else node_table
        self.func_graph = nx.DiGraph() if func_graph is None else func_graph
        self.function_node_table = (
            FunctionNodeTable.from_node_table(self.node_table)
            if function_node_table is None
            else function_node_table
        )

    @classmethod
    def from_perf_script_data(
        cls,
        perf_script: PerfScriptData,
        pids: list | None = None,
        cpus: list | None = None,
        filter_none: bool = False,
        gen_epm: bool = False,
        perf_buildid: Optional[PerfBuildidData] = None,
        source_file_prefix: Optional[str] = None,
    ):
        """
        Creates a CallGraph object from performance script data.

        Args:
            perf_script (PerfScriptData): The performance script data.
            pid (int, optional): The process ID. Defaults to None.
            cpu (int, optional): The CPU ID. Defaults to None.
            filter_none (bool): Whether to filter out unknown functions. Defaults to False.
            gen_epm (bool): Whether to generate EPM. Defaults to False.
            perf_buildid (PerfBuildidData, optional): The buildid data. Defaults to None.
            source_file_prefix (str, optional): The prefix of source file. Defaults to None.

        Returns:
            CallGraph: The CallGraph object created from the performance script data.
        """
        if pids is not None:
            perf_script = perf_script.filter_by_pids(pids=pids)
        if cpus is not None:
            perf_script = perf_script.filter_by_cpus(cpus=cpus)

        node_table = NodeTable.from_perf_script_data(perf_script)
        block_graph = nx.DiGraph()

        buildid_list = {}
        if perf_buildid:
            buildid_list = perf_buildid.buildid_lists
        func_table = FunctionNodeTable.from_node_table(
            node_table,
            gen_epm=gen_epm,
            buildid_list=buildid_list,
            source_file_prefix=source_file_prefix,
        )
        func_graph = nx.DiGraph()

        for block in perf_script.blocks:
            calls = block.calls
            for i in range(1, len(calls)):
                if filter_none:
                    for j in range(i - 1, -1, -1):
                        if node_table[calls[j].addr].get_function_name() != "[unknown]":
                            caller = calls[j].addr
                            break
                else:
                    caller = calls[i - 1].addr
                callee = calls[i].addr

                if filter_none and (
                    node_table[caller].get_function_name() == "[unknown]"
                    or node_table[callee].get_function_name() == "[unknown]"
                ):
                    continue
                block_graph.add_edge(node_table[callee], node_table[caller], weight=1)
                k_caller = f"{node_table[caller].get_function_name()} {node_table[caller].caller}"
                k_callee = f"{node_table[callee].get_function_name()} {node_table[callee].caller}"
                func_caller = func_table[k_caller]
                func_callee = func_table[k_callee]
                if func_graph.has_edge(func_callee, func_caller):
                    func_graph[func_callee][func_caller]["weight"] += 1
                else:
                    func_graph.add_edge(
                        func_callee,
                        func_caller,
                        weight=1,
                    )

        return cls(
            block_graph=block_graph,
            node_table=node_table,
            func_graph=func_graph,
            function_node_table=func_table,
        )

    def simple_groups(
        self,
        fig_path: str = "simple_groups.png",
        cluster_info_path: str = "simple_groups_cluster.txt",
        supergraph_layout_scale: int = 50,
        supergraph_layout_seed: int = 429,
        supergraph_layout_k: Optional[float] = None,
        supergraph_layout_iters: int = 50,
        nodegroup_layout_scale: int = 1,
        nodegroup_layout_seed: int = 1430,
        nodegroup_layout_k: Optional[float] = None,
        nodegroup_layout_iters: int = 50,
    ):
        """
        Simply group graph with its module name

        Args:
            fig_path (str, optional): Save figure to file. Defaults to "simple_groups.png".
            cluster_info_path (str, optional): Save raw cluster data to file. Defaults to "simple_groups_cluster.txt".
            supergraph_layout_scale (int, optional): The whole graph's layout scale param. Defaults to 50.
            supergraph_layout_seed (int, optional): The whole graph's layout seed param. Defaults to 429.
            supergraph_layout_k (Optional[float], optional): The whole graph's layout k param. Defaults to None.
            supergraph_layout_iters (int, optional): The whole graph's layout iters param. Defaults to 100.
            nodegroup_layout_scale (int, optional): Each node group graph's layout scale param. Defaults to 20.
            nodegroup_layout_seed (int, optional): The node group graph's layout seed param. Defaults to 1430.
            nodegroup_layout_k (Optional[float], optional): The node group graph's layout k param. Defaults to None.
            nodegroup_layout_iters (int, optional): The node group graph's layout iters param. Defaults to 100.

        Examples:
        >>> from pipa.service.call_graph import CallGraph
        >>> from pipa.parser.perf_script_call import PerfScriptData
        >>> data = PerfScriptData.from_file("perf.script")
        >>> cfg = CallGraph.from_perf_script_data(data)
        >>> cfg.simple_groups()
        """
        G = self.func_graph
        nodes = G.nodes

        # create groups
        attrs_groups = defaultdict(lambda: [])
        for node in nodes:
            attr = f"{node.module_name}"
            attrs_groups[attr].append(node)
        attrs_to_cluster = {attr: idx for idx, attr in enumerate(attrs_groups.keys())}

        # assign cluster & Combine Data
        _clusters = defaultdict(lambda: {"cycles": 0, "insts": 0, "funcs": []})
        for node, node_v in nodes.items():
            attr = f"{node.module_name}"
            _cluster = attrs_to_cluster[attr]
            node_v["cluster"] = _cluster
            _clusters[_cluster]["cycles"] += node.get_cycles()
            _clusters[_cluster]["insts"] += node.get_instructions()
            _clusters[_cluster]["funcs"].extend(node.nodes)  # type: ignore
            # for sub_node in node.nodes:
        with open(cluster_info_path, "w") as file:
            json.dump(_clusters, file, cls=ClusterEncoder, indent=4)

        # use viridis colors for mapping
        color_map = plt.get_cmap("viridis", len(attrs_groups))

        # set color for the group results
        node_colors = [color_map(node_v["cluster"]) for node_v in nodes.values()]
        for i, node_v in enumerate(nodes.values()):
            node_v["color"] = node_colors[i]

        # fetch each node's position
        # group nodes
        pos = {}
        node_groups = [frozenset(nodes) for nodes in attrs_groups.values()]
        superpos = nx.spring_layout(
            G,
            scale=supergraph_layout_scale,
            seed=supergraph_layout_seed,
            k=supergraph_layout_k,
            iterations=supergraph_layout_iters,
        )
        centers = list(superpos.values())
        for center, comm in zip(centers, node_groups):
            pos.update(
                nx.spring_layout(
                    nx.subgraph(G, comm),
                    center=center,
                    scale=nodegroup_layout_scale,
                    seed=nodegroup_layout_seed,
                    k=nodegroup_layout_k,
                    iterations=nodegroup_layout_iters,
                )
            )

        # specify node's name
        node_names = {}
        for node in nodes:
            node_names[node] = (
                f"{node}\ncycles: {node.get_cycles()}\ninsts: {node.get_instructions()}"
            )

        # print fig
        self.show(
            graph="func_graph",
            node_names=node_names,
            pos=pos,
            fig_path=fig_path,
            node_color=node_colors,
            node_groups=node_groups,
        )

    def show(
        self,
        pos: Optional[Mapping] = None,
        node_names: Optional[Mapping] = None,
        graph: Literal["block_graph", "func_graph"] = "func_graph",
        layout_scale: int = 3,
        fig_path: Optional[str] = None,
        node_color: str | list = "skyblue",
        fig_size: tuple[int, int] = (100, 100),
        node_size: int = 700,
        font_size: int = 12,
        font_weight: Literal["normal", "bold"] = "normal",
        node_groups: Optional[list] = None,
    ):
        """
        Displays the call graph.

        Args:
            pos (Optional[Mapping], optional): The graph nodes' position data, if None is passed will calculated using default spring_layout. Defaults to None.
            node_names (Optional[Mapping], optional): The graph nodes' name. Defaults to None.
            graph (Literal[&quot;block_graph&quot;, &quot;func_graph&quot;], optional): Which type of graph to show. Defaults to "func_graph".
            layout_scale (int, optional): default spring_layout's scale param. Defaults to 3.
            fig_path (Optional[str], optional): The path to save the call graph figure. Defaults to None.
            node_color (str | list, optional): The graph nodes' color. Defaults to "skyblue". Can be a list
            fig_size (tuple[int, int], optional): The figure size. Defaults to (100, 100).
            node_size (int, optional): The graph nodes' size. Defaults to 700.
            font_size (int, optional): The font size in figure. Defaults to 12.
            font_weight (Literal[&quot;normal&quot;, &quot;bold&quot;], optional): The font weight in figure. Defaults to "normal".
            node_groups (Optional[list], optional): Use node groups to draw nodes separately. Defaults to None.
        """
        G = self.__getattribute__(graph)
        plt.figure(figsize=fig_size)

        # require node positions
        if not pos:
            pos = nx.spring_layout(G, scale=layout_scale)

        # draw edges' label
        edge_labels = nx.get_edge_attributes(G, "weight")
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)

        # draw nodes
        if node_groups:
            for nodes in node_groups:
                nx.draw_networkx_nodes(G, pos=pos, nodelist=nodes)

        # draw graph with additional information
        nx.draw(
            G,
            pos,
            labels=node_names,
            with_labels=True,
            node_size=node_size,
            node_color=node_color,
            font_size=font_size,
            font_weight=font_weight,
        )

        # print figure
        plt.tight_layout()
        if fig_path:
            plt.savefig(fig_path)
        plt.show()

    def save_dot(self, dot_path: str):
        """
        Saves the call graph dot file.

        Args:
            dot_path (str): The path to save the call graph.

        Returns:
            None
        """
        write_dot(self.block_graph, dot_path)



class sadf:
    """sadf 数据模型基类，封装所有监控指标的表头结构"""

    class CpuUsage:
        """CPU 使用率统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC）
        cpu: str  # CPU 编号（-1 表示总计）
        user_percent: float  # 用户态使用率 (%usr)
        nice_percent: float  # Nice 进程使用率 (%nice)
        system_percent: float  # 内核态使用率 (%sys)
        iowait_percent: float  # I/O 等待时间 (%iowait)
        steal_percent: float  # 被其他虚拟机偷走的时间 (%steal)
        irq_percent: float  # 硬中断时间 (%irq)
        soft_percent: float  # 软中断时间 (%soft)
        guest_percent: float  # 虚拟机运行时间 (%guest)
        gnice_percent: float  # 调整过优先级的虚拟机时间 (%gnice)
        idle_percent: float  # 空闲时间 (%idle)

    class ProcessStats:
        """进程与上下文切换统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC）
        processes_per_second: float  # 每秒进程调度数 (proc/s)
        context_switches_per_second: float  # 每秒上下文切换次数 (cswch/s)

    class Interrupts:
        """中断统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC）
        interrupt_id: str  # 中断编号 (INTR)
        interrupts_per_second: float  # 每秒中断次数 (intr/s)

    class SwapStats:
        """交换分区统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        pswpin_per_second: float  # 每秒从交换区读取页数 (pswpin/s)
        pswpout_per_second: float  # 每秒写入交换区页数 (pswpout/s)

    class MemoryPageStats:
        """内存页统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        pgpgin_per_second: float  # 每秒从磁盘读取页数 (pgpgin/s)
        pgpgout_per_second: float  # 每秒写入磁盘页数 (pgpgout/s)
        faults_per_second: float  # 每秒页面错误数 (fault/s)
        majflt_per_second: float  # 每秒重大页面错误数 (majflt/s)
        pgfree_per_second: float  # 每秒释放页数 (pgfree/s)
        pgscank_per_second: float  # 每秒扫描页数 (pgscank/s)
        pgscand_per_second: float  # 每秒直接扫描页数 (pgscand/s)
        pgsteal_per_second: float  # 每秒偷取页数 (pgsteal/s)
        vmeff_percent: float  # 虚拟内存效率 (%vmeff)

    class DiskIoStats:
        """磁盘 I/O 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        tps: float  # 每秒传输次数 (tps)
        rtps: float  # 每秒读取次数 (rtps)
        wtps: float  # 每秒写入次数 (wtps)
        dtps: float  # 每秒删除次数 (dtps)
        bread_per_second: float  # 每秒读取块数 (bread/s)
        bwrtn_per_second: float  # 每秒写入块数 (bwrtn/s)
        bdscd_per_second: float  # 每秒丢弃块数 (bdscd/s)

    class MemoryUsage:
        """内存使用统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        kbmemfree: int  # 空闲内存大小 (kbmemfree)
        kbavail: int  # 可用内存大小 (kbavail)
        kbmemused: int  # 已用内存大小 (kbmemused)
        memused_percent: float  # 内存使用率 (%memused)
        kbbuffers: int  # 缓冲区大小 (kbbuffers)
        kbcached: int  # 缓存大小 (kbcached)
        kbcommit: int  # 提交的内存量 (kbcommit)
        commit_percent: float  # 提交内存使用率 (%commit)
        kbactive: int  # 活跃内存大小 (kbactive)
        kbinact: int  # 非活跃内存大小 (kbinact)
        kbdirty: int  # 脏页大小 (kbdirty)
        kbanonpg: int  # 匿名页大小 (kbanonpg)
        kbslab: int  # Slab 分配器大小 (kbslab)
        kbkstack: int  # 内核栈大小 (kbkstack)
        kbpgtbl: int  # 页表大小 (kbpgtbl)
        kbvmused: int  # 虚拟内存使用量 (kbvmused)

    class SwapUsage:
        """交换分区使用统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        kbswpfree: int  # 空闲交换分区大小 (kbswpfree)
        kbswpused: int  # 已用交换分区大小 (kbswpused)
        swpused_percent: float  # 交换分区使用率 (%swpused)
        kbswpcad: int  # 压缩交换分区大小 (kbswpcad)
        swpcad_percent: float  # 压缩交换分区使用率 (%swpcad)

    class HugePages:
        """大页内存统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        kbhugfree: int  # 空闲大页内存大小 (kbhugfree)
        kbhugused: int  # 已用大页内存大小 (kbhugused)
        hugused_percent: float  # 大页内存使用率 (%hugused)
        kbhugrsvd: int  # 预留大页内存大小 (kbhugrsvd)
        kbhugsurp: int  # 超出预留的大页内存大小 (kbhugsurp)

    class FileDescriptors:
        """文件描述符统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        dentunusd: int  # 未使用的目录项 (dentunusd)
        file_nr: int  # 文件描述符总数 (file-nr)
        inode_nr: int  # inode 数量 (inode-nr)
        pty_nr: int  # 伪终端数量 (pty-nr)

    class LoadAverage:
        """系统负载统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        runq_sz: int  # 运行队列长度 (runq-sz)
        plist_sz: int  # 进程列表大小 (plist-sz)
        ldavg_1: float  # 1 分钟负载平均值 (ldavg-1)
        ldavg_5: float  # 5 分钟负载平均值 (ldavg-5)
        ldavg_15: float  # 15 分钟负载平均值 (ldavg-15)
        blocked: int  # 被阻塞的进程数 (blocked)

    class DiskDeviceStats:
        """磁盘设备统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        dev: str  # 设备名称 (DEV)
        tps: float  # 每秒传输次数 (tps)
        rkB_per_second: float  # 每秒读取千字节数 (rkB/s)
        wkB_per_second: float  # 每秒写入千字节数 (wkB/s)
        dkB_per_second: float  # 每秒删除千字节数 (dkB/s)
        areq_sz: float  # 平均请求大小 (areq-sz)
        aqu_sz: float  # 平均队列长度 (aqu-sz)
        await_time: float  # 平均等待时间 (await)
        util_percent: float  # 设备利用率 (%util)

    class NetworkInterfaceStats:
        """网络接口统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        iface: str  # 接口名称 (IFACE)
        rxpck_per_second: float  # 每秒接收数据包数 (rxpck/s)
        txpck_per_second: float  # 每秒发送数据包数 (txpck/s)
        rxkB_per_second: float  # 每秒接收千字节数 (rxkB/s)
        txkB_per_second: float  # 每秒发送千字节数 (txkB/s)
        rxcmp_per_second: float  # 每秒接收压缩包数 (rxcmp/s)
        txcmp_per_second: float  # 每秒发送压缩包数 (txcmp/s)
        rxmcst_per_second: float  # 每秒接收多播包数 (rxmcst/s)
        ifutil_percent: float  # 接口利用率 (%ifutil)

    class NetworkErrors:
        """网络错误统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        iface: str  # 接口名称 (IFACE)
        rxerr_per_second: float  # 每秒接收错误数 (rxerr/s)
        txerr_per_second: float  # 每秒发送错误数 (txerr/s)
        coll_per_second: float  # 每秒冲突数 (coll/s)
        rxdrop_per_second: float  # 每秒接收丢弃数 (rxdrop/s)
        txdrop_per_second: float  # 每秒发送丢弃数 (txdrop/s)
        txcarr_per_second: float  # 每秒发送载波错误数 (txcarr/s)
        rxfram_per_second: float  # 每秒接收帧错误数 (rxfram/s)
        rxfifo_per_second: float  # 每秒接收 FIFO 错误数 (rxfifo/s)
        txfifo_per_second: float  # 每秒发送 FIFO 错误数 (txfifo/s)

    class NfsStats:
        """NFS 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        call_per_second: float  # 每秒调用次数 (call/s)
        retrans_per_second: float  # 每秒重传次数 (retrans/s)
        read_per_second: float  # 每秒读取次数 (read/s)
        write_per_second: float  # 每秒写入次数 (write/s)
        access_per_second: float  # 每秒访问次数 (access/s)
        getatt_per_second: float  # 每秒获取属性次数 (getatt/s)

    class SocketsStats:
        """套接字统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        scall_per_second: float  # 每秒系统调用次数 (scall/s)
        badcall_per_second: float  # 每秒错误调用次数 (badcall/s)
        packet_per_second: float  # 每秒数据包数 (packet/s)
        udp_per_second: float  # 每秒 UDP 包数 (udp/s)
        tcp_per_second: float  # 每秒 TCP 包数 (tcp/s)
        hit_per_second: float  # 每秒命中次数 (hit/s)
        miss_per_second: float  # 每秒未命中次数 (miss/s)
        sread_per_second: float  # 每秒读取次数 (sread/s)
        swrite_per_second: float  # 每秒写入次数 (swrite/s)
        saccess_per_second: float  # 每秒访问次数 (saccess/s)
        sgetatt_per_second: float  # 每秒获取属性次数 (sgetatt/s)

    class TcpUdpStats:
        """TCP/UDP 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        totsck: int  # 总套接字数 (totsck)
        tcpsck: int  # TCP 套接字数 (tcpsck)
        udpsck: int  # UDP 套接字数 (udpsck)
        rawsck: int  # RAW 套接字数 (rawsck)
        ip_frag: int  # IP 分片数 (ip-frag)
        tcp_tw: int  # TCP TIME-WAIT 状态数 (tcp-tw)

    class IcmpStats:
        """ICMP 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        irec_per_second: float  # 每秒接收 ICMP 包数 (irec/s)
        fwddgm_per_second: float  # 每秒转发数据报数 (fwddgm/s)
        idel_per_second: float  # 每秒删除数据报数 (idel/s)
        orq_per_second: float  # 每秒输出请求数 (orq/s)
        asmrq_per_second: float  # 每秒组装请求数 (asmrq/s)
        asmok_per_second: float  # 每秒成功组装数 (asmok/s)
        fragok_per_second: float  # 每秒成功分片数 (fragok/s)
        fragcrt_per_second: float  # 每秒创建分片数 (fragcrt/s)

    class IcmpErrors:
        """ICMP 错误统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        ihdrerr_per_second: float  # 每秒头部错误数 (ihdrerr/s)
        iadrerr_per_second: float  # 每秒地址错误数 (iadrerr/s)
        iukwnpr_per_second: float  # 每秒未知协议错误数 (iukwnpr/s)
        idisc_per_second: float  # 每秒丢弃数 (idisc/s)
        odisc_per_second: float  # 每秒输出丢弃数 (odisc/s)
        onort_per_second: float  # 每秒无路由错误数 (onort/s)
        asmf_per_second: float  # 每秒组装失败数 (asmf/s)
        fragf_per_second: float  # 每秒分片失败数 (fragf/s)

    class IcmpMessages:
        """ICMP 消息统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        imsg_per_second: float  # 每秒接收消息数 (imsg/s)
        omsg_per_second: float  # 每秒发送消息数 (omsg/s)
        iech_per_second: float  # 每秒接收回声请求数 (iech/s)
        iechr_per_second: float  # 每秒接收回声应答数 (iechr/s)
        oech_per_second: float  # 每秒发送回声请求数 (oech/s)
        oechr_per_second: float  # 每秒发送回声应答数 (oechr/s)
        itm_per_second: float  # 每秒接收时间戳请求数 (itm/s)
        itmr_per_second: float  # 每秒接收时间戳应答数 (itmr/s)
        otm_per_second: float  # 每秒发送时间戳请求数 (otm/s)
        otmr_per_second: float  # 每秒发送时间戳应答数 (otmr/s)
        iadrmk_per_second: float  # 每秒接收地址掩码请求数 (iadrmk/s)
        iadrmkr_per_second: float  # 每秒接收地址掩码应答数 (iadrmkr/s)
        oadrmk_per_second: float  # 每秒发送地址掩码请求数 (oadrmk/s)
        oadrmkr_per_second: float  # 每秒发送地址掩码应答数 (oadrmkr/s)

    class IcmpRedirects:
        """ICMP 重定向统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        ierr_per_second: float  # 每秒错误数 (ierr/s)
        oerr_per_second: float  # 每秒输出错误数 (oerr/s)
        idstunr_per_second: float  # 每秒目标不可达错误数 (idstunr/s)
        odstunr_per_second: float  # 每秒输出目标不可达错误数 (odstunr/s)
        itmex_per_second: float  # 每秒超时错误数 (itmex/s)
        otmex_per_second: float  # 每秒输出超时错误数 (otmex/s)
        iparmpb_per_second: float  # 每秒参数错误数 (iparmpb/s)
        oparmpb_per_second: float  # 每秒输出参数错误数 (oparmpb/s)
        isrcq_per_second: float  # 每秒源抑制请求数 (isrcq/s)
        osrcq_per_second: float  # 每秒输出源抑制请求数 (osrcq/s)
        iredir_per_second: float  # 每秒重定向请求数 (iredir/s)
        oredir_per_second: float  # 每秒输出重定向请求数 (oredir/s)

    class TcpStats:
        """TCP 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        active_per_second: float  # 每秒主动连接数 (active/s)
        passive_per_second: float  # 每秒被动连接数 (passive/s)
        iseg_per_second: float  # 每秒接收段数 (iseg/s)
        oseg_per_second: float  # 每秒发送段数 (oseg/s)

    class TcpErrors:
        """TCP 错误统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        atmptf_per_second: float  # 每秒尝试失败连接数 (atmptf/s)
        estres_per_second: float  # 每秒建立连接数 (estres/s)
        retrans_per_second: float  # 每秒重传次数 (retrans/s)
        isegerr_per_second: float  # 每秒接收错误段数 (isegerr/s)
        orsts_per_second: float  # 每秒输出重置数 (orsts/s)

    class UdpStats:
        """UDP 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        idgm_per_second: float  # 每秒接收数据报数 (idgm/s)
        odgm_per_second: float  # 每秒发送数据报数 (odgm/s)
        noport_per_second: float  # 每秒无端口数据报数 (noport/s)
        idgmerr_per_second: float  # 每秒接收错误数据报数 (idgmerr/s)

    class IPv6Stats:
        """IPv6 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        tcp6sck: int  # IPv6 TCP 套接字数 (tcp6sck)
        udp6sck: int  # IPv6 UDP 套接字数 (udp6sck)
        raw6sck: int  # IPv6 RAW 套接字数 (raw6sck)
        ip6_frag: int  # IPv6 分片数 (ip6-frag)

    class IPv6IcmpStats:
        """IPv6 ICMP 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        irec6_per_second: float  # 每秒接收 ICMPv6 包数 (irec6/s)
        fwddgm6_per_second: float  # 每秒转发数据报数 (fwddgm6/s)
        idel6_per_second: float  # 每秒删除数据报数 (idel6/s)
        orq6_per_second: float  # 每秒输出请求数 (orq6/s)
        asmrq6_per_second: float  # 每秒组装请求数 (asmrq6/s)
        asmok6_per_second: float  # 每秒成功组装数 (asmok6/s)
        imcpck6_per_second: float  # 每秒接收 MIB ICMP 包数 (imcpck6/s)
        omcpck6_per_second: float  # 每秒发送 MIB ICMP 包数 (omcpck6/s)
        fragok6_per_second: float  # 每秒成功分片数 (fragok6/s)
        fragcr6_per_second: float  # 每秒创建分片数 (fragcr6/s)

    class IPv6IcmpErrors:
        """IPv6 ICMP 错误统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        ihdrer6_per_second: float  # 每秒头部错误数 (ihdrer6/s)
        iadrer6_per_second: float  # 每秒地址错误数 (iadrer6/s)
        iukwnp6_per_second: float  # 每秒未知协议错误数 (iukwnp6/s)
        i2big6_per_second: float  # 每秒过大包错误数 (i2big6/s)
        idisc6_per_second: float  # 每秒丢弃数 (idisc6/s)
        odisc6_per_second: float  # 每秒输出丢弃数 (odisc6/s)
        inort6_per_second: float  # 每秒无路由错误数 (inort6/s)
        onort6_per_second: float  # 每秒输出无路由错误数 (onort6/s)
        asmf6_per_second: float  # 每秒组装失败数 (asmf6/s)
        fragf6_per_second: float  # 每秒分片失败数 (fragf6/s)
        itrpck6_per_second: float  # 每秒陷阱包数 (itrpck6/s)

    class IPv6IcmpMessages:
        """IPv6 ICMP 消息统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        imsg6_per_second: float  # 每秒接收消息数 (imsg6/s)
        omsg6_per_second: float  # 每秒发送消息数 (omsg6/s)
        iech6_per_second: float  # 每秒接收回声请求数 (iech6/s)
        iechr6_per_second: float  # 每秒接收回声应答数 (iechr6/s)
        oechr6_per_second: float  # 每秒发送回声应答数 (oechr6/s)
        igmbq6_per_second: float  # 每秒接收组播查询数 (igmbq6/s)
        igmbr6_per_second: float  # 每秒接收组播报告数 (igmbr6/s)
        ogmbr6_per_second: float  # 每秒发送组播报告数 (ogmbr6/s)
        igmbrd6_per_second: float  # 每秒接收组播离开数 (igmbrd6/s)
        ogmbrd6_per_second: float  # 每秒发送组播离开数 (ogmbrd6/s)
        irtsol6_per_second: float  # 每秒接收路由器求解数 (irtsol6/s)
        ortsol6_per_second: float  # 每秒发送路由器求解数 (ortsol6/s)
        irtad6_per_second: float  # 每秒接收路由器通告数 (irtad6/s)
        inbsol6_per_second: float  # 每秒接收邻居求解数 (inbsol6/s)
        onbsol6_per_second: float  # 每秒发送邻居求解数 (onbsol6/s)
        inbad6_per_second: float  # 每秒接收邻居通告数 (inbad6/s)
        onbad6_per_second: float  # 每秒发送邻居通告数 (onbad6/s)

    class IPv6IcmpRedirects:
        """IPv6 ICMP 重定向统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        ierr6_per_second: float  # 每秒错误数 (ierr6/s)
        idtunr6_per_second: float  # 每秒目标不可达错误数 (idtunr6/s)
        odtunr6_per_second: float  # 每秒输出目标不可达错误数 (odtunr6/s)
        itmex6_per_second: float  # 每秒超时错误数 (itmex6/s)
        otmex6_per_second: float  # 每秒输出超时错误数 (otmex6/s)
        iprmpb6_per_second: float  # 每秒参数错误数 (iprmpb6/s)
        oprmpb6_per_second: float  # 每秒输出参数错误数 (oprmpb6/s)
        iredir6_per_second: float  # 每秒重定向请求数 (iredir6/s)
        oredir6_per_second: float  # 每秒输出重定向请求数 (oredir6/s)
        ipck2b6_per_second: float  # 每秒包过大错误数 (ipck2b6/s)
        opck2b6_per_second: float  # 每秒输出包过大错误数 (opck2b6/s)

    class IPv6UdpStats:
        """IPv6 UDP 统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        idgm6_per_second: float  # 每秒接收数据报数 (idgm6/s)
        odgm6_per_second: float  # 每秒发送数据报数 (odgm6/s)
        noport6_per_second: float  # 每秒无端口数据报数 (noport6/s)
        idgmer6_per_second: float  # 每秒接收错误数据报数 (idgmer6/s)

    class CpuFrequency:
        """CPU 频率统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        cpu: str  # CPU 编号 (CPU)
        mhz: float  # CPU 当前频率 (MHz)

    class CpuTemperature:
        """CPU 温度统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        temp: str  # 温度编号 (TEMP)
        device: str  # 温度传感器设备 (DEVICE)
        degc: float  # 温度 (degC)
        temp_percent: float  # 温度百分比 (%temp)

    class UsbDevices:
        """USB 设备统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        manufact: str  # 制造商 (manufact)
        product: str  # 产品 (product)
        bus: str  # 总线 (BUS)
        idvendor: str  # 供应商 ID (idvendor)
        idprod: str  # 产品 ID (idprod)
        maxpower: str  # 最大功率 (maxpower)

    class FilesystemUsage:
        """文件系统使用统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        filesystem: str  # 文件系统 (FILESYSTEM)
        mbfsfree: int  # 空闲空间大小 (MBfsfree)
        mbfsused: int  # 已用空间大小 (MBfsused)
        fsused_percent: float  # 空间使用率 (%fsused)
        ufsused_percent: float  # 用户空间使用率 (%ufsused)
        ifree: int  # 空闲 inode 数 (Ifree)
        iused: int  # 已用 inode 数 (Iused)
        iused_percent: float  # inode 使用率 (%Iused)

    class CpuQueueStats:
        """CPU 队列统计"""

        hostname: str  # 主机名
        interval: str  # 监控间隔时间（秒）
        timestamp: str  # 时间戳（UTC)
        cpu: str  # CPU 编号 (CPU)
        total_per_second: float  # 每秒总流量 (total/s)
        dropd_per_second: float  # 每秒丢包数 (dropd/s)
        squeezd_per_second: float  # 每秒压缩包数 (squeezd/s)
        rx_rps_per_second: float  # 每秒 RPS 接收数 (rx_rps/s)
        flw_lim_per_second: float  # 每秒流量限制数 (flw_lim/s)


def parse_sadf_data(file_path):
    """
    解析 sadf 生成的文本数据，返回结构化的字典列表。

    Args:
        file_path (str): 要解析的文件路径。

    Returns:
        list: 每个元素是一个字典，包含 'header'（表头）和 'rows'（数据行）。
    """
    result = []
    current_block = None

    with open(file_path, "r") as f:
        for line in f:
            line = line.strip()
            if not line:
                continue

            # 判断是否为表头行（以 # 开头）
            if line.startswith("#"):
                if current_block:
                    result.append(current_block)  # 保存上一个数据块
                # 提取表头字段
                header = [field.strip() for field in line[1:].split(";")]
                current_block = {"header": header, "rows": []}
            else:
                # 解析数据行
                fields = [field.strip() for field in line.split(";")]
                if len(fields) != len(current_block["header"]):
                    raise ValueError(
                        f"字段数不匹配: 表头有 {len(current_block['header'])} 字段，数据行有 {len(fields)} 字段"
                    )
                current_block["rows"].append(dict(zip(current_block["header"], fields)))

        # 添加最后一个块
        if current_block:
            result.append(current_block)

    return result


if __name__ == "__main__":
    data_blocks = parse_sadf_data("sadf.csv")

    for block in data_blocks:
        print("表头:", block["header"])
        print("前3行数据:")
        for row in block["rows"][:3]:
            print(row)
        print("\n---")

    print(len(data_blocks))  # 输出数据块的数量



from .pipad.pipad_client import PIPADClient
from pipa.common.utils import get_timestamp
from pipa.common.logger import logger
import questionary
import os
import argparse


def download(ip, port, table_name: str, file_format: str, dirs: str = "."):
    file_rcv: bytes = PIPADClient.download_full_table(ip, port, table_name, file_format)
    table_path = os.path.join(dirs, f"{table_name}_{get_timestamp()}.{file_format}")
    try:
        with open(table_path, "wb") as f:
            f.write(file_rcv)
        print(f"Table {table_name} downloaded successfully to {table_path}")
    except Exception as e:
        logger.error(f"Failed to download table {table_name}: {e}")


def prompt_user_for_details():
    ip = questionary.text("What's PIPAD IP address?").ask()
    port = questionary.text("What's PIPAD port?").ask()
    table_name = questionary.text("What's table name?").ask()
    file_format = questionary.select(
        "Which file format do you want to download?",
        choices=["xlsx", "csv"],
    ).ask()
    return ip, port, table_name, file_format


def main():
    parser = argparse.ArgumentParser(description="Download table from PIPAD")
    parser.add_argument("--ip", type=str, help="PIPAD IP address")
    parser.add_argument("--port", type=str, help="PIPAD port")
    parser.add_argument("--table", type=str, help="Table name")
    parser.add_argument(
        "--format", type=str, choices=["xlsx", "csv"], help="File format"
    )
    args = parser.parse_args()

    if args.ip and args.port and args.table and args.format:
        ip, port, table_name, file_format = (
            args.ip,
            args.port,
            args.table,
            args.format,
        )
    else:
        ip, port, table_name, file_format = prompt_user_for_details()
    download(ip, port, table_name, file_format)



import pytest
from unittest.mock import patch, mock_open
from pipa.common.utils import FileFormat, check_file_format


@patch(
    "builtins.open", new_callable=mock_open, read_data=b"\xfd\x37\x7a\x58\x5a\x00"
)  # xz magic number
def test_check_file_format_xz(mock_open):
    assert check_file_format(file="test.xz") == FileFormat.xz
    mock_open.assert_called_with("test.xz", "rb")


@patch("builtins.open", new_callable=mock_open, read_data=b"BZh")  # bz2 magic number
def test_check_file_format_bzip2(mock_open):
    assert check_file_format(file="test.xz") == FileFormat.bzip2
    mock_open.assert_called_with("test.xz", "rb")


@patch(
    "builtins.open", new_callable=mock_open, read_data=b"\x7f\x45\x4c\x46"
)  # elf magic number
def test_check_file_format_elf(mock_open):
    assert check_file_format(file="test.xz") == FileFormat.elf
    mock_open.assert_called_with("test.xz", "rb")


@patch(
    "builtins.open", new_callable=mock_open, read_data=b"RandomBytes"
)  # other file format
def test_check_file_format_other(mock_open):
    assert check_file_format(file="test.xz") == FileFormat.other
    mock_open.assert_called_with("test.xz", "rb")


if __name__ == "__main__":  # pragma: no cover
    pytest.main([__file__])



from typing import List, Optional
import plotly.graph_objects as go


def make_single_plot(
    scatters: List[go.Scatter],
    title: str,
    xaxis_title: str,
    yaxis_title: str,
    show: bool = True,
    write_to_html: Optional[str] = None,
) -> go.Figure:
    """Make single plotly graph based on the list of scatters.

    Args:
        scatters (List[go.Scatter]): Scsatters to plot.
        title (str): Title of the plot.
        xaxis_title (str): Title of the x-axis.
        yaxis_title (str): Title of the y-axis.
        show (bool, optional): Whether to show the fig directly. Defaults to True.
        write_to_html (Optional[str], optional): If specified, will write fig to the path with html format. Defaults to None.

    Returns:
        go.Figure: The plotly figure.
    """
    fig = go.Figure()
    for s in scatters:
        fig.add_trace(s)
    fig.update_layout(
        title=title,
        xaxis_title=xaxis_title,
        yaxis_title=yaxis_title,
        hovermode="closest",
    )
    if show:
        fig.show()
    if write_to_html:
        fig.write_html(write_to_html)
    return fig



import unittest
import os
import pandas as pd
from unittest.mock import patch, MagicMock
from pipa.common.export import SQLiteConnector, export_dataframe_to_csv


class TestSQLiteConnector(unittest.TestCase):
    def setUp(self):
        self.db_path = ":memory:"
        self.connector = SQLiteConnector(self.db_path)
        self.test_df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})

    @patch("sqlite3.connect")
    def test_fetch_table_as_dataframe(self, mock_connect):
        mock_conn = MagicMock()
        mock_connect.return_value = mock_conn
        mock_conn.__enter__.return_value.cursor.return_value.execute.return_value = None
        mock_conn.__enter__.return_value.cursor.return_value.fetchall.return_value = [
            (1, 4),
            (2, 5),
            (3, 6),
        ]
        mock_conn.__enter__.return_value.cursor.return_value.description = [
            ("A",),
            ("B",),
        ]

        result = self.connector.fetch_table_as_dataframe("test_table")
        self.assertTrue(isinstance(result, pd.DataFrame))
        self.assertEqual(result.shape, (3, 2))

    @patch("sqlite3.connect")
    def test_execute_query(self, mock_connect):
        mock_conn = MagicMock()
        mock_connect.return_value = mock_conn
        mock_conn.__enter__.return_value.cursor.return_value.execute.return_value = None
        mock_conn.__enter__.return_value.cursor.return_value.fetchall.return_value = [
            (1, 4),
            (2, 5),
            (3, 6),
        ]
        mock_conn.__enter__.return_value.cursor.return_value.description = [
            ("A",),
            ("B",),
        ]

        result = self.connector.execute_query("SELECT * FROM test_table")
        self.assertTrue(isinstance(result, pd.DataFrame))
        self.assertEqual(result.shape, (3, 2))

    @patch("sqlite3.connect")
    def test_export_table_to_csv(self, mock_connect):
        mock_conn = MagicMock()
        mock_connect.return_value = mock_conn
        mock_conn.__enter__.return_value.cursor.return_value.execute.return_value = None
        mock_conn.__enter__.return_value.cursor.return_value.fetchall.return_value = [
            (1, 4),
            (2, 5),
            (3, 6),
        ]
        mock_conn.__enter__.return_value.cursor.return_value.description = [
            ("A",),
            ("B",),
        ]

        output_file = "test_output.csv"
        self.connector.export_table_to_csv("test_table", output_file)
        self.assertTrue(os.path.exists(output_file))
        os.remove(output_file)


class TestExportDataframeToCSV(unittest.TestCase):
    def test_export_dataframe_to_csv(self):
        @export_dataframe_to_csv("test_output.csv")
        def dummy_function():
            return pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})

        dummy_function()
        self.assertTrue(os.path.exists("test_output.csv"))
        df = pd.read_csv("test_output.csv")
        self.assertEqual(df.shape, (3, 2))
        self.assertTrue(df.equals(pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})))
        os.remove("test_output.csv")


if __name__ == "__main__":
    unittest.main()
